# copyright 2003-2016 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of CubicWeb.
#
# CubicWeb is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# CubicWeb is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with CubicWeb.  If not, see <http://www.gnu.org/licenses/>.
"""The edit controller, automatically handling entity form submitting"""

from warnings import warn
from collections import defaultdict

from datetime import datetime

from six import text_type

from logilab.common.graph import ordered_nodes

from rql.utils import rqlvar_maker

from cubicweb import _, ValidationError, UnknownEid
from cubicweb.view import EntityAdapter
from cubicweb.predicates import is_instance
from cubicweb.web import RequestError, NothingToEdit, ProcessFormError
from cubicweb.web.views import basecontrollers, autoform


class IEditControlAdapter(EntityAdapter):
    __regid__ = 'IEditControl'
    __select__ = is_instance('Any')

    def __init__(self, _cw, **kwargs):
        if self.__class__ is not IEditControlAdapter:
            warn('[3.14] IEditControlAdapter is deprecated, override EditController'
                 ' using match_edited_type or match_form_id selectors for example.',
                 DeprecationWarning)
        super(IEditControlAdapter, self).__init__(_cw, **kwargs)

    def after_deletion_path(self):
        """return (path, parameters) which should be used as redirect
        information when this entity is being deleted
        """
        parent = self.entity.cw_adapt_to('IBreadCrumbs').parent_entity()
        if parent is not None:
            return parent.rest_path(), {}
        return str(self.entity.e_schema).lower(), {}

    def pre_web_edit(self):
        """callback called by the web editcontroller when an entity will be
        created/modified, to let a chance to do some entity specific stuff.

        Do nothing by default.
        """
        pass


def valerror_eid(eid):
    try:
        return int(eid)
    except (ValueError, TypeError):
        return eid


class RqlQuery(object):
    def __init__(self):
        self.edited = []
        self.restrictions = []
        self.kwargs = {}
        self.canceled = False

    def __repr__(self):
        return ('Query <edited=%r restrictions=%r kwargs=%r>' % (
            self.edited, self.restrictions, self.kwargs))

    def insert_query(self, etype):
        assert not self.canceled
        if self.edited:
            rql = 'INSERT %s X: %s' % (etype, ','.join(self.edited))
        else:
            rql = 'INSERT %s X' % etype
        if self.restrictions:
            rql += ' WHERE %s' % ','.join(self.restrictions)
        return rql

    def update_query(self, eid):
        assert not self.canceled
        varmaker = rqlvar_maker()
        var = next(varmaker)
        while var in self.kwargs:
            var = next(varmaker)
        rql = 'SET %s WHERE X eid %%(%s)s' % (','.join(self.edited), var)
        if self.restrictions:
            rql += ', %s' % ','.join(self.restrictions)
        self.kwargs[var] = eid
        return rql

    def set_attribute(self, attr, value):
        self.kwargs[attr] = value
        self.edited.append('X %s %%(%s)s' % (attr, attr))

    def set_inlined(self, relation, value):
        self.kwargs[relation] = value
        self.edited.append('X %s %s' % (relation, relation.upper()))
        self.restrictions.append('%s eid %%(%s)s' % (relation.upper(), relation))


class EditController(basecontrollers.ViewController):
    __regid__ = 'edit'

    def publish(self, rset=None):
        """edit / create / copy / delete entity / relations"""
        for key in self._cw.form:
            # There should be 0 or 1 action
            if key.startswith('__action_'):
                cbname = key[1:]
                try:
                    callback = getattr(self, cbname)
                except AttributeError:
                    raise RequestError(self._cw._('invalid action %r' % key))
                else:
                    return callback()
        self._default_publish()
        self.reset()

    def _ordered_formparams(self):
        """ Return form parameters dictionaries for each edited entity.

        We ensure that entities can be created in this order accounting for
        mandatory inlined relations.
        """
        req = self._cw
        graph = {}
        get_rschema = self._cw.vreg.schema.rschema
        # minparams = 2, because at least __type and eid are needed
        values_by_eid = dict((eid, req.extract_entity_params(eid, minparams=2))
                             for eid in req.edited_eids())
        # iterate over all the edited entities
        for eid, values in values_by_eid.items():
            # add eid to the dependency graph
            graph.setdefault(eid, set())
            # search entity's edited fields for mandatory inlined relation
            for param in values['_cw_entity_fields'].split(','):
                try:
                    rtype, role = param.split('-')
                except ValueError:
                    # e.g. param='__type'
                    continue
                rschema = get_rschema(rtype)
                if rschema.inlined:
                    for target in rschema.targets(values['__type'], role):
                        rdef = rschema.role_rdef(values['__type'], target, role)
                        # if cardinality is 1 and if the target entity is being
                        # simultaneously edited, the current entity must be
                        # created before the target one
                        if rdef.cardinality[0 if role == 'subject' else 1] == '1':
                            # use .get since param may be unspecified (though it will usually lead
                            # to a validation error later)
                            target_eid = values.get(param)
                            if target_eid in values_by_eid:
                                # add dependency from the target entity to the
                                # current one
                                if role == 'object':
                                    graph.setdefault(target_eid, set()).add(eid)
                                else:
                                    graph.setdefault(eid, set()).add(target_eid)
                                break
        for eid in reversed(ordered_nodes(graph)):
            yield values_by_eid[eid]

    def _default_publish(self):
        req = self._cw
        self.errors = []
        self.relations_rql = []
        form = req.form
        # so we're able to know the main entity from the repository side
        if '__maineid' in form:
            req.transaction_data['__maineid'] = form['__maineid']
        # no specific action, generic edition
        self._to_create = req.data['eidmap'] = {}
        # those three data variables are used to handle relation from/to entities
        # which doesn't exist at time where the entity is edited and that
        # deserves special treatment
        req.data['pending_inlined'] = defaultdict(set)
        req.data['pending_others'] = set()
        req.data['pending_composite_delete'] = set()
        req.data['pending_values'] = dict()
        try:
            for formparams in self._ordered_formparams():
                self.edit_entity(formparams)
        except (RequestError, NothingToEdit) as ex:
            if '__linkto' in req.form and 'eid' in req.form:
                self.execute_linkto()
            elif '__delete' not in req.form:
                raise ValidationError(None, {None: text_type(ex)})
        # all pending inlined relations to newly created entities have been
        # treated now (pop to ensure there are no attempt to add new ones)
        pending_inlined = req.data.pop('pending_inlined')
        assert not pending_inlined, pending_inlined
        pending_values = req.data.pop('pending_values')
        # handle all other remaining relations now
        while req.data['pending_others']:
            form_, field = req.data['pending_others'].pop()
            # attempt to retrieve values and original values if they have already gone through
            # handle_formfield (may not if there has been some not yet known eid at the first
            # processing round). In the later case we've to go back through handle_formfield.
            try:
                values, origvalues = pending_values.pop((form_, field))
            except KeyError:
                self.handle_formfield(form_, field)
            else:
                self.handle_relation(form_, field, values, origvalues)
        assert not pending_values, 'unexpected remaining pending values %s' % pending_values
        del req.data['pending_others']
        # then execute rql to set all relations
        for querydef in self.relations_rql:
            self._cw.execute(*querydef)
        # delete pending composite
        for entity in req.data['pending_composite_delete']:
            entity.cw_delete()
        # XXX this processes *all* pending operations of *all* entities
        if '__delete' in req.form:
            todelete = req.list_form_param('__delete', req.form, pop=True)
            if todelete:
                autoform.delete_relations(self._cw, todelete)
        self._cw.remove_pending_operations()
        if self.errors:
            errors = dict((f.name, text_type(ex)) for f, ex in self.errors)
            raise ValidationError(valerror_eid(form.get('__maineid')), errors)

    def _insert_entity(self, etype, eid, rqlquery):
        rql = rqlquery.insert_query(etype)
        try:
            entity = self._cw.execute(rql, rqlquery.kwargs).get_entity(0, 0)
            neweid = entity.eid
        except ValidationError as ex:
            self._to_create[eid] = ex.entity
            if self._cw.ajax_request:  # XXX (syt) why?
                ex.entity = eid
            raise
        self._to_create[eid] = neweid
        return neweid

    def _update_entity(self, eid, rqlquery):
        self._cw.execute(rqlquery.update_query(eid), rqlquery.kwargs)

    def edit_entity(self, formparams, multiple=False):
        """edit / create / copy an entity and return its eid"""
        req = self._cw
        etype = formparams['__type']
        entity = req.vreg['etypes'].etype_class(etype)(req)
        entity.eid = valerror_eid(formparams['eid'])
        is_main_entity = req.form.get('__maineid') == formparams['eid']
        # let a chance to do some entity specific stuff
        entity.cw_adapt_to('IEditControl').pre_web_edit()
        # create a rql query from parameters
        rqlquery = RqlQuery()
        # process inlined relations at the same time as attributes
        # this will generate less rql queries and might be useful in
        # a few dark corners
        if is_main_entity:
            formid = req.form.get('__form_id', 'edition')
        else:
            # XXX inlined forms formid should be saved in a different formparams entry
            # inbetween, use cubicweb standard formid for inlined forms
            formid = 'edition'
        form = req.vreg['forms'].select(formid, req, entity=entity)
        eid = form.actual_eid(entity.eid)
        editedfields = formparams['_cw_entity_fields']
        form.formvalues = {}  # init fields value cache
        for field in form.iter_modified_fields(editedfields, entity):
            self.handle_formfield(form, field, rqlquery)
        # if there are some inlined field which were waiting for this entity's
        # creation, add relevant data to the rqlquery
        for form_, field in req.data['pending_inlined'].pop(entity.eid, ()):
            rqlquery.set_inlined(field.name, form_.edited_entity.eid)
        if not rqlquery.canceled:
            if self.errors:
                errors = dict((f.role_name(), text_type(ex)) for f, ex in self.errors)
                raise ValidationError(valerror_eid(entity.eid), errors)
            if eid is None:  # creation or copy
                entity.eid = eid = self._insert_entity(etype, formparams['eid'], rqlquery)
            elif rqlquery.edited:  # edition of an existant entity
                self.check_concurrent_edition(formparams, eid)
                self._update_entity(eid, rqlquery)
        else:
            self.errors = []
        if is_main_entity:
            self.notify_edited(entity)
        if '__delete' in formparams:
            # XXX deprecate?
            todelete = req.list_form_param('__delete', formparams, pop=True)
            autoform.delete_relations(req, todelete)
        if '__cloned_eid' in formparams:
            entity.copy_relations(int(formparams['__cloned_eid']))
        if is_main_entity:  # only execute linkto for the main entity
            self.execute_linkto(entity.eid)
        return eid

    def handle_formfield(self, form, field, rqlquery=None):
        entity = form.edited_entity
        eschema = entity.e_schema
        try:
            for field, value in field.process_posted(form):
                if not ((field.role == 'subject' and field.name in eschema.subjrels)
                        or
                        (field.role == 'object' and field.name in eschema.objrels)):
                    continue

                rschema = self._cw.vreg.schema.rschema(field.name)
                if rschema.final:
                    rqlquery.set_attribute(field.name, value)
                    continue

                if entity.has_eid():
                    origvalues = set(row[0] for row in entity.related(field.name, field.role).rows)
                else:
                    origvalues = set()
                if value is None or value == origvalues:
                    continue  # not edited / not modified / to do later

                unlinked_eids = origvalues - value

                if unlinked_eids:
                    # Special handling of composite relation removal
                    self.handle_composite_removal(
                        form, field, unlinked_eids, value, rqlquery)

                if rschema.inlined and rqlquery is not None and field.role == 'subject':
                    self.handle_inlined_relation(form, field, value, origvalues, rqlquery)
                elif form.edited_entity.has_eid():
                    self.handle_relation(form, field, value, origvalues)
                else:
                    form._cw.data['pending_others'].add((form, field))
                    form._cw.data['pending_values'][(form, field)] = (value, origvalues)

        except ProcessFormError as exc:
            self.errors.append((field, exc))

    def handle_composite_removal(self, form, field,
                                 removed_values, new_values, rqlquery):
        """
        In EditController-handled forms, when the user removes a composite
        relation, it triggers the removal of the related entity in the
        composite. This is where this happens.

        See for instance test_subject_subentity_removal in
        web/test/unittest_application.py.
        """
        rschema = self._cw.vreg.schema.rschema(field.name)
        new_value_etypes = set()
        # the user could have included nonexisting eids in the POST; don't crash.
        for eid in new_values:
            try:
                new_value_etypes.add(self._cw.entity_from_eid(eid).cw_etype)
            except UnknownEid:
                continue
        for unlinked_eid in removed_values:
            unlinked_entity = self._cw.entity_from_eid(unlinked_eid)
            rdef = rschema.role_rdef(form.edited_entity.cw_etype,
                                     unlinked_entity.cw_etype,
                                     field.role)
            if rdef.composite is not None:
                if rdef.composite == field.role:
                    to_be_removed = unlinked_entity
                else:
                    if unlinked_entity.cw_etype in new_value_etypes:
                        # This is a same-rdef re-parenting: do not remove the entity
                        continue
                    to_be_removed = form.edited_entity
                    self.info('Edition of %s is cancelled (deletion requested)',
                              to_be_removed)
                    rqlquery.canceled = True
                self.info('Scheduling removal of %s as composite relation '
                          '%s was removed', to_be_removed, rdef)
                form._cw.data['pending_composite_delete'].add(to_be_removed)

    def handle_inlined_relation(self, form, field, values, origvalues, rqlquery):
        """handle edition for the (rschema, x) relation of the given entity
        """
        if values:
            rqlquery.set_inlined(field.name, next(iter(values)))
        elif form.edited_entity.has_eid():
            self.handle_relation(form, field, values, origvalues)

    def handle_relation(self, form, field, values, origvalues):
        """handle edition for the (rschema, x) relation of the given entity
        """
        rschema = self._cw.vreg.schema.rschema(field.name)
        if field.role == 'subject':
            subjvar, objvar = 'X', 'Y'
        else:
            subjvar, objvar = 'Y', 'X'
        eid = form.edited_entity.eid
        if field.role == 'object' or not rschema.inlined or not values:
            # this is not an inlined relation or no values specified,
            # explicty remove relations
            rql = 'DELETE %s %s %s WHERE X eid %%(x)s, Y eid %%(y)s' % (
                subjvar, rschema, objvar)
            for reid in origvalues.difference(values):
                self.relations_rql.append((rql, {'x': eid, 'y': reid}))
        seteids = values.difference(origvalues)
        if seteids:
            rql = 'SET %s %s %s WHERE X eid %%(x)s, Y eid %%(y)s' % (
                subjvar, rschema, objvar)
            for reid in seteids:
                self.relations_rql.append((rql, {'x': eid, 'y': reid}))

    def delete_entities(self, eidtypes):
        """delete entities from the repository"""
        redirect_info = set()
        eidtypes = tuple(eidtypes)
        for eid, etype in eidtypes:
            entity = self._cw.entity_from_eid(eid, etype)
            path, params = entity.cw_adapt_to('IEditControl').after_deletion_path()
            redirect_info.add((path, tuple(params.items())))
            entity.cw_delete()
        if len(redirect_info) > 1:
            # In the face of ambiguity, refuse the temptation to guess.
            self._after_deletion_path = 'view', ()
        else:
            self._after_deletion_path = next(iter(redirect_info))
        if len(eidtypes) > 1:
            self._cw.set_message(self._cw._('entities deleted'))
        else:
            self._cw.set_message(self._cw._('entity deleted'))

    def check_concurrent_edition(self, formparams, eid):
        req = self._cw
        try:
            form_ts = datetime.utcfromtimestamp(float(formparams['__form_generation_time']))
        except KeyError:
            # Backward and tests compatibility : if no timestamp consider edition OK
            return
        if req.execute("Any X WHERE X modification_date > %(fts)s, X eid %(eid)s",
                       {'eid': eid, 'fts': form_ts}):
            # We only mark the message for translation but the actual
            # translation will be handled by the Validation mechanism...
            msg = _("Entity %(eid)s has changed since you started to edit it."
                    " Reload the page and reapply your changes.")
            # ... this is why we pass the formats' dict as a third argument.
            raise ValidationError(eid, {None: msg}, {'eid': eid})

    def _action_apply(self):
        self._default_publish()
        self.reset()

    def _action_delete(self):
        self.delete_entities(self._cw.edited_eids(withtype=True))
        return self.reset()



from base64 import b64decode


# before removing extid, ensure it's coherent with cwuri
for eid, etype, encoded_extid in sql(
        "SELECT eid, type, extid FROM entities, cw_CWSource "
        "WHERE cw_CWSource.cw_name=entities.asource AND cw_CWSource.cw_type='ldapfeed'"):
    sql('UPDATE cw_{} SET cw_cwuri=%(cwuri)s WHERE cw_eid=%(eid)s'.format(etype),
        {'eid': eid, 'cwuri': b64decode(encoded_extid)})

sql('DROP TABLE moved_entities')
sql('ALTER TABLE entities DROP COLUMN asource')
sql('ALTER TABLE entities DROP COLUMN extid')
sql('DROP INDEX entities_type_idx')

# force cw_schema deletion before CWSourceSchemaConfig to avoid nasty bug
drop_relation_type('cw_schema')
drop_entity_type('CWSourceSchemaConfig')

# Check the CW versions and add the entity only if needed ?
add_entity_type('CWSession')
rql('DELETE CWProperty X WHERE X pkey "system.version.pyramid"',
    ask_confirm=False)



# copyright 2003-2011 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of CubicWeb.
#
# CubicWeb is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# CubicWeb is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with CubicWeb.  If not, see <http://www.gnu.org/licenses/>.
"""This module provides some generic components to navigate in the web
application.

Pagination
----------

Several implementations for large result set pagination are provided:

.. autoclass:: PageNavigation
.. autoclass:: PageNavigationSelect
.. autoclass:: SortedNavigation

Pagination will appear when needed according to the `page-size` ui property.

This module monkey-patch the :func:`paginate` function to the base :class:`View`
class, so that you can ask pagination explicitly on every result-set based views.

.. autofunction:: paginate


Previous / next navigation
--------------------------

An adapter and its related component for the somewhat usal "previous / next"
navigation are provided.

  .. autoclass:: IPrevNextAdapter
  .. autoclass:: NextPrevNavigationComponent
"""


from cubicweb import _

from datetime import datetime

from six import text_type

from rql.nodes import VariableRef, Constant

from logilab.mtconverter import xml_escape
from logilab.common.deprecation import deprecated

from cubicweb.predicates import paginated_rset, sorted_rset, adaptable
from cubicweb.uilib import cut
from cubicweb.view import EntityAdapter
from cubicweb.web.component import EmptyComponent, EntityCtxComponent, NavigationComponent


class PageNavigation(NavigationComponent):
    """The default pagination component: display link to pages where each pages
    is identified by the item number of its first and last elements.
    """
    def call(self):
        """displays a resultset by page"""
        params = dict(self._cw.form)
        self.clean_params(params)
        basepath = self._cw.relative_path(includeparams=False)
        self.w(u'<div class="pagination">')
        self.w(self.previous_link(basepath, params))
        self.w(u'[&#160;%s&#160;]' %
               u'&#160;| '.join(self.iter_page_links(basepath, params)))
        self.w(u'&#160;&#160;%s' % self.next_link(basepath, params))
        self.w(u'</div>')

    def index_display(self, start, stop):
        return u'%s - %s' % (start+1, stop+1)

    def iter_page_links(self, basepath, params):
        rset = self.cw_rset
        page_size = self.page_size
        start = 0
        while start < rset.rowcount:
            stop = min(start + page_size - 1, rset.rowcount - 1)
            yield self.page_link(basepath, params, start, stop,
                                 self.index_display(start, stop))
            start = stop + 1


class PageNavigationSelect(PageNavigation):
    """This pagination component displays a result-set by page as
    :class:`PageNavigation` but in a <select>, which is better when there are a
    lot of results.

    By default it will be selected when there are more than 4 pages to be
    displayed.
    """
    __select__ = paginated_rset(4)

    page_link_templ = u'<option value="%s" title="%s">%s</option>'
    selected_page_link_templ = u'<option value="%s" selected="selected" title="%s">%s</option>'
    def call(self):
        params = dict(self._cw.form)
        self.clean_params(params)
        basepath = self._cw.relative_path(includeparams=False)
        w = self.w
        w(u'<div class="pagination">')
        w(self.previous_link(basepath, params))
        w(u'<select onchange="javascript: document.location=this.options[this.selectedIndex].value">')
        for option in self.iter_page_links(basepath, params):
            w(option)
        w(u'</select>')
        w(u'&#160;&#160;%s' % self.next_link(basepath, params))
        w(u'</div>')


class SortedNavigation(NavigationComponent):
    """This pagination component will be selected by default if there are less
    than 4 pages and if the result set is sorted.

    Displayed links to navigate accross pages of a result set are done according
    to the first variable on which the sort is done, and looks like:

        [ana - cro] | [cro - ghe] | ... | [tim - zou]

    You may want to override this component to customize display in some cases.

    .. automethod:: sort_on
    .. automethod:: display_func
    .. automethod:: format_link_content
    .. automethod:: write_links

    Below an example from the tracker cube:

    .. sourcecode:: python

      class TicketsNavigation(navigation.SortedNavigation):
          __select__ = (navigation.SortedNavigation.__select__
                        & ~paginated_rset(4) & is_instance('Ticket'))
          def sort_on(self):
              col, attrname = super(TicketsNavigation, self).sort_on()
              if col == 6:
                  # sort on state, we don't want that
                  return None, None
              return col, attrname

    The idea is that in trackers'ticket tables, result set is first ordered on
    ticket's state while this doesn't make any sense in the navigation. So we
    override :meth:`sort_on` so that if we detect such sorting, we disable the
    feature to go back to item number in the pagination.

    Also notice the `~paginated_rset(4)` in the selector so that if there are
    more than 4 pages to display, :class:`PageNavigationSelect` will still be
    selected.
    """
    __select__ = paginated_rset() & sorted_rset()

    # number of considered chars to build page links
    nb_chars = 5

    def call(self):
        # attrname = the name of attribute according to which the sort
        # is done if any
        col, attrname = self.sort_on()
        index_display = self.display_func(self.cw_rset, col, attrname)
        basepath = self._cw.relative_path(includeparams=False)
        params = dict(self._cw.form)
        self.clean_params(params)
        blocklist = []
        start = 0
        total = self.cw_rset.rowcount
        while start < total:
            stop = min(start + self.page_size - 1, total - 1)
            cell = self.format_link_content(index_display(start), index_display(stop))
            blocklist.append(self.page_link(basepath, params, start, stop, cell))
            start = stop + 1
        self.write_links(basepath, params, blocklist)

    def display_func(self, rset, col, attrname):
        """Return a function that will be called with a row number as argument
        and should return a string to use as link for it.
        """
        if attrname is not None:
            def index_display(row):
                if not rset[row][col]: # outer join
                    return u''
                entity = rset.get_entity(row, col)
                return entity.printable_value(attrname, format='text/plain')
        elif col is None: # smart links disabled.
            def index_display(row):
                return text_type(row)
        elif self._cw.vreg.schema.eschema(rset.description[0][col]).final:
            def index_display(row):
                return text_type(rset[row][col])
        else:
            def index_display(row):
                return rset.get_entity(row, col).view('text')
        return index_display

    def sort_on(self):
        """Return entity column number / attr name to use for nice display by
        inspecting the rset'syntax tree.
        """
        rschema = self._cw.vreg.schema.rschema
        for sorterm in self.cw_rset.syntax_tree().children[0].orderby:
            if isinstance(sorterm.term, Constant):
                col = sorterm.term.value - 1
                return col, None
            var = sorterm.term.get_nodes(VariableRef)[0].variable
            col = None
            for ref in var.references():
                rel = ref.relation()
                if rel is None:
                    continue
                attrname = rel.r_type
                if attrname in ('is', 'has_text'):
                    continue
                if not rschema(attrname).final:
                    col = var.selected_index()
                    attrname = None
                if col is None:
                    # final relation or not selected non final relation
                    if var is rel.children[0]:
                        relvar = rel.children[1].children[0].get_nodes(VariableRef)[0]
                    else:
                        relvar = rel.children[0].variable
                    col = relvar.selected_index()
                if col is not None:
                    break
            else:
                # no relation but maybe usable anyway if selected
                col = var.selected_index()
                attrname = None
            if col is not None:
                # if column type is date[time], set proper 'nb_chars'
                if var.stinfo['possibletypes'] & frozenset(('TZDatetime', 'Datetime',
                                                            'Date')):
                    self.nb_chars = len(self._cw.format_date(datetime.today()))
                return col, attrname
        # nothing usable found, use the first column
        return 0, None

    def format_link_content(self, startstr, stopstr):
        """Return text for a page link, where `startstr` and `stopstr` are the
        text for the lower/upper boundaries of the page.

        By default text are stripped down to :attr:`nb_chars` characters.
        """
        text = u'%s - %s' % (startstr.lower()[:self.nb_chars],
                             stopstr.lower()[:self.nb_chars])
        return xml_escape(text)

    def write_links(self, basepath, params, blocklist):
        """Return HTML for the whole navigation: `blocklist` is a list of HTML
        snippets for each page, `basepath` and `params` will be necessary to
        build previous/next links.
        """
        self.w(u'<div class="pagination">')
        self.w(u'%s&#160;' % self.previous_link(basepath, params))
        self.w(u'[&#160;%s&#160;]' % u'&#160;| '.join(blocklist))
        self.w(u'&#160;%s' % self.next_link(basepath, params))
        self.w(u'</div>')


def do_paginate(view, rset=None, w=None, show_all_option=True, page_size=None):
    """write pages index in w stream (default to view.w) and then limit the
    result set (default to view.rset) to the currently displayed page if we're
    not explicitly told to display everything (by setting __force_display in
    req.form)
    """
    req = view._cw
    if rset is None:
        rset = view.cw_rset
    if w is None:
        w = view.w
    nav = req.vreg['components'].select_or_none(
        'navigation', req, rset=rset, page_size=page_size, view=view)
    if nav:
        if w is None:
            w = view.w
        if req.form.get('__force_display'):
            nav.render_link_back_to_pagination(w=w)
        else:
            # get boundaries before component rendering
            start, stop = nav.page_boundaries()
            nav.render(w=w)
            if show_all_option:
                nav.render_link_display_all(w=w)
            rset.limit(offset=start, limit=stop-start, inplace=True)


def paginate(view, show_all_option=True, w=None, page_size=None, rset=None):
    """paginate results if the view is paginable
    """
    if view.paginable:
        do_paginate(view, rset, w, show_all_option, page_size)

# monkey patch base View class to add a .paginate([...])
# method to be called to write pages index in the view and then limit the result
# set to the current page
from cubicweb.view import View
View.do_paginate = do_paginate
View.paginate = paginate
View.handle_pagination = False



class IPrevNextAdapter(EntityAdapter):
    """Interface for entities which can be linked to a previous and/or next
    entity

    .. automethod:: next_entity
    .. automethod:: previous_entity
    """
    __needs_bw_compat__ = True
    __regid__ = 'IPrevNext'
    __abstract__ = True

    def next_entity(self):
        """return the 'next' entity"""
        raise NotImplementedError

    def previous_entity(self):
        """return the 'previous' entity"""
        raise NotImplementedError


class NextPrevNavigationComponent(EntityCtxComponent):
    """Entities adaptable to the 'IPrevNext' should have this component
    automatically displayed. You may want to override this component to have a
    different look and feel.
    """

    __regid__ = 'prevnext'
    # register msg not generated since no entity implements IPrevNext in cubicweb
    # itself
    help = _('ctxcomponents_prevnext_description')
    __select__ = EntityCtxComponent.__select__ & adaptable('IPrevNext')
    context = 'navbottom'
    order = 10

    @property
    def prev_icon(self):
        return '<img src="%s" alt="%s" />' % (
            xml_escape(self._cw.data_url('go_prev.png')), self._cw._('previous page'))

    @property
    def next_icon(self):
        return '<img src="%s" alt="%s" />' % (
            xml_escape(self._cw.data_url('go_next.png')), self._cw._('next page'))

    def init_rendering(self):
        adapter = self.entity.cw_adapt_to('IPrevNext')
        self.previous = adapter.previous_entity()
        self.next = adapter.next_entity()
        if not (self.previous or self.next):
            raise EmptyComponent()

    def render_body(self, w):
        w(u'<div class="prevnext">')
        self.prevnext(w)
        w(u'</div>')
        w(u'<div class="clear"></div>')

    def prevnext(self, w):
        if self.previous:
            self.prevnext_entity(w, self.previous, 'prev')
        if self.next:
            self.prevnext_entity(w, self.next, 'next')

    def prevnext_entity(self, w, entity, type):
        textsize = self._cw.property_value('navigation.short-line-size')
        content = xml_escape(cut(entity.dc_title(), textsize))
        if type == 'prev':
            title = self._cw._('i18nprevnext_previous')
            icon = self.prev_icon
            cssclass = u'previousEntity left'
            content = icon + '&#160;&#160;' + content
        else:
            title = self._cw._('i18nprevnext_next')
            icon = self.next_icon
            cssclass = u'nextEntity right'
            content = content + '&#160;&#160;' + icon
        self.prevnext_div(w, type, cssclass, entity.absolute_url(),
                          title, content)

    def prevnext_div(self, w, type, cssclass, url, title, content):
        w(u'<div class="%s">' % cssclass)
        w(u'<a href="%s" title="%s">%s</a>' % (xml_escape(url),
                                               xml_escape(title),
                                               content))
        w(u'</div>')
        self._cw.html_headers.add_raw('<link rel="%s" href="%s" />' % (
              type, xml_escape(url)))



# copyright 2003-2013 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of CubicWeb.
#
# CubicWeb is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# CubicWeb is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with CubicWeb.  If not, see <http://www.gnu.org/licenses/>.
"""Mass mailing handling: send mail to entities adaptable to IEmailable"""

try:
    from cubes.massmailing.views import (SendEmailAction,
                                         recipient_vocabulary,
                                         MassMailingForm,
                                         MassMailingFormRenderer,
                                         MassMailingFormView,
                                         SendMailController)


    from logilab.common.deprecation import class_moved, moved

    msg = '[3.17] cubicweb.web.views.massmailing moved to cubes.massmailing.views'
    SendEmailAction = class_moved(SendEmailAction, message=msg)
    recipient_vocabulary = moved('cubes.massmailing.views', 'recipient_vocabulary')
    MassMailingForm = class_moved(MassMailingForm, message=msg)
    MassMailingFormRenderer = class_moved(MassMailingFormRenderer, message=msg)
    MassMailingFormView = class_moved(MassMailingFormView, message=msg)
    SendMailController = class_moved(SendMailController, message=msg)
except ImportError:
    from cubicweb.web import LOGGER
    LOGGER.warning('[3.17] massmailing extracted to cube massmailing that was not found. try installing it.')



sync_schema_props_perms('EmailAddress')

for source in rql('CWSource X WHERE X type "ldapuser"').entities():
    config = source.dictconfig
    host = config.pop('host', u'ldap')
    protocol = config.pop('protocol', u'ldap')
    source.cw_set(url=u'%s://%s' % (protocol, host))
    source.update_config(skip_unknown=True, **config)

commit()



# copyright 2003-2010 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of CubicWeb.
#
# CubicWeb is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# CubicWeb is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with CubicWeb.  If not, see <http://www.gnu.org/licenses/>.
"""

"""
from logilab.common.testlib import TestCase, unittest_main

from cubicweb.web.views.apacherewrite import *

class ApacheURLRewriteTC(TestCase):

    def test(self):
        class MyAppRules(ApacheURLRewrite):
            rules = [
                RewriteCond('logilab\.fr', match='host',
                            rules=[('/(.*)', r'http://www.logilab.fr/\1')],
                            action='redirect'),
                RewriteCond('(www)\.logilab\.fr', match='host', action='stop'),
                RewriteCond('/(data|json)/', match='path', action='stop'),
                RewriteCond('(?P<cat>.*)\.logilab\.fr', match='host',
                            rules=[('/(.*)', r'/m_%(cat)s/\1')]),
                ]
        urlrewriter = MyAppRules()
        req = None # not used in the above rules, so keep a simple TestCase here
        try:
            urlrewriter.rewrite('logilab.fr', '/whatever', req)
            self.fail('redirect exception expected')
        except Redirect as ex:
            self.assertEqual(ex.location, 'http://www.logilab.fr/whatever')
        self.assertEqual(urlrewriter.rewrite('www.logilab.fr', '/whatever', req),
                          '/whatever')
        self.assertEqual(urlrewriter.rewrite('www.logilab.fr', '/json/bla', req),
                          '/json/bla')
        self.assertEqual(urlrewriter.rewrite('abcd.logilab.fr', '/json/bla', req),
                          '/json/bla')
        self.assertEqual(urlrewriter.rewrite('abcd.logilab.fr', '/data/bla', req),
                          '/data/bla')
        self.assertEqual(urlrewriter.rewrite('abcd.logilab.fr', '/whatever', req),
                          '/m_abcd/whatever')
        self.assertEqual(urlrewriter.rewrite('abcd.fr', '/whatever', req),
                          '/whatever')


if __name__ == '__main__':
    unittest_main()



# copyright 2003-2014 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of CubicWeb.
#
# CubicWeb is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# CubicWeb is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with CubicWeb.  If not, see <http://www.gnu.org/licenses/>.

from six import text_type

from logilab.common import tempattr

from cubicweb.devtools.testlib import CubicWebTC
from cubicweb.devtools.fake import FakeRequest

from cubicweb.web.views.urlrewrite import (SimpleReqRewriter, SchemaBasedRewriter,
                                           rgx, rgx_action)


class UrlRewriteTC(CubicWebTC):

    def test_auto_extend_rules(self):
        class Rewriter(SimpleReqRewriter):
            rules = [
                ('foo', dict(rql='Foo F')),
                ('/index', dict(vid='index2')),
                ]
        rules = []
        for pattern, values in Rewriter.rules:
            if hasattr(pattern, 'pattern'):
                pattern = pattern.pattern
            rules.append((pattern, values))
        self.assertListEqual(rules, [
            ('foo' , dict(rql='Foo F')),
            ('/index' , dict(vid='index2')),
            ('/_', dict(vid='manage')),
            ('/_registry', dict(vid='registry')),
            ('/schema', dict(vid='schema')),
            ('/myprefs', dict(vid='propertiesform')),
            ('/siteconfig', dict(vid='systempropertiesform')),
            ('/siteinfo', dict(vid='siteinfo')),
            ('/manage', dict(vid='manage')),
            ('/notfound', dict(vid='404')),
            ('/error', dict(vid='error')),
            ('/sparql', dict(vid='sparql')),
            ('/processinfo', dict(vid='processinfo')),
            ('/cwuser$', {'vid': 'cw.users-and-groups-management',
                          'tab': 'cw_users_management'}),
            ('/cwgroup$', {'vid': 'cw.users-and-groups-management',
                           'tab': 'cw_groups_management'}),
            ('/cwsource$', {'vid': 'cw.sources-management'}),
            ('/schema/([^/]+?)/?$', {'rql': r'Any X WHERE X is CWEType, X name "\1"', 'vid': 'primary'}),
            ('/add/([^/]+?)/?$' , dict(vid='creation', etype=r'\1')),
            ('/doc/images/(.+?)/?$', dict(fid='\\1', vid='wdocimages')),
            ('/doc/?$', dict(fid='main', vid='wdoc')),
            ('/doc/(.+?)/?$', dict(fid='\\1', vid='wdoc')),
            # now in SchemaBasedRewriter
            #('/search/(.+)$', dict(rql=r'Any X WHERE X has_text "\1"')),
            ])


    def test_no_extend_rules(self):
        class Rewriter(SimpleReqRewriter):
            ignore_baseclass_rules = True
            rules = [
                ('foo', dict(rql='Foo F')),
                ('/index', dict(vid='index2')),
                ]
        self.assertListEqual(Rewriter.rules, [
            ('foo' , dict(rql='Foo F')),
            ('/index' , dict(vid='index2')),
            ])

    def test_basic_transformation(self):
        """test simple string-based rewrite"""
        req = FakeRequest()
        rewriter = SimpleReqRewriter(req)
        self.assertRaises(KeyError, rewriter.rewrite, req, '/view?vid=whatever')
        self.assertEqual(req.form, {})
        rewriter.rewrite(req, '/index')
        self.assertEqual(req.form, {'vid' : "index"})

    def test_regexp_transformation(self):
        """test regexp-based rewrite"""
        req = FakeRequest()
        rewriter = SimpleReqRewriter(req)
        rewriter.rewrite(req, '/add/Task')
        self.assertEqual(req.form, {'vid' : "creation", 'etype' : "Task"})
        req = FakeRequest()
        rewriter.rewrite(req, '/add/Task/')
        self.assertEqual(req.form, {'vid' : "creation", 'etype' : "Task"})

    def test_inheritance(self):
        BaseTransition = self.vreg['etypes'].etype_class('BaseTransition')
        with self.admin_access.web_request() as req:
            x = req.create_entity('WorkflowTransition', name=u'test')
            ctrlid, rset = self.app.url_resolver.process(req, 'basetransition/%s' % x.eid)
            self.assertEqual(ctrlid, 'view')
            self.assertEqual(x.eid, rset[0][0])
            # cw_rest_attr_info is cached but clear_cache doesn't like cached class
            # method
            del BaseTransition._cw_rest_attr_info_cache_
            try:
                with tempattr(BaseTransition, 'rest_attr', 'name'):

                    ctrlid, rset = self.app.url_resolver.process(req, 'basetransition/%s' % x.name)
                    self.assertEqual(ctrlid, 'view')
                    self.assertEqual(x.eid, rset[0][0])
            finally:
                del BaseTransition._cw_rest_attr_info_cache_



class RgxActionRewriteTC(CubicWebTC):

    def setup_database(self):
        with self.admin_access.repo_cnx() as cnx:
            p1 = self.create_user(cnx, u'user1')
            p1.cw_set(firstname=u'joe', surname=u'Dalton')
            p2 = self.create_user(cnx, u'user2')
            p2.cw_set(firstname=u'jack', surname=u'Dalton')
            self.p1eid = p1.eid
            cnx.commit()

    def test_rgx_action_with_transforms(self):
        class TestSchemaBasedRewriter(SchemaBasedRewriter):
            rules = [
                (rgx('/(?P<sn>\w+)/(?P<fn>\w+)'),
                 rgx_action(r'Any X WHERE X surname %(sn)s, '
                            'X firstname %(fn)s',
                            argsgroups=('sn', 'fn'),
                            transforms={'sn' : text_type.capitalize,
                                        'fn' : text_type.lower,})),
                ]
        with self.admin_access.web_request() as req:
            rewriter = TestSchemaBasedRewriter(req)
            _pmid, rset = rewriter.rewrite(req, u'/DaLToN/JoE')
            self.assertEqual(len(rset), 1)
            self.assertEqual(rset[0][0], self.p1eid)
            self.assertEqual(rset.description[0][0], 'CWUser')

    def test_inheritance_precedence(self):
        RQL1 = 'Any C WHERE C is CWEType'
        RQL2 = 'Any C WHERE C is CWUser'

        class BaseRewriter(SchemaBasedRewriter):
            rules = [
               (rgx('/collector(.*)'),
                rgx_action(rql=RQL1,
                    form=dict(vid='baseindex')),
                ),
                ]
        class Rewriter(BaseRewriter):
            rules = [
               (rgx('/collector/something(/?)'),
                rgx_action(rql=RQL2,
                    form=dict(vid='index')),
                ),
                ]

        with self.admin_access.web_request() as req:
            rewriter = Rewriter(req)
            _pmid, rset = rewriter.rewrite(req, '/collector')
            self.assertEqual(rset.rql, RQL1)
            self.assertEqual(req.form, {'vid' : "baseindex"})
            _pmid, rset = rewriter.rewrite(req, '/collector/something')
            self.assertEqual(rset.rql, RQL2)
            self.assertEqual(req.form, {'vid' : "index"})
            _pmid, rset = rewriter.rewrite(req, '/collector/something/')
            self.assertEqual(req.form, {'vid' : "index"})
            self.assertEqual(rset.rql, RQL2)
            _pmid, rset = rewriter.rewrite(req, '/collector/somethingelse/')
            self.assertEqual(rset.rql, RQL1)
            self.assertEqual(req.form, {'vid' : "baseindex"})

    def test_inheritance_precedence_same_rgx(self):
        RQL1 = 'Any C WHERE C is CWEType'
        RQL2 = 'Any C WHERE C is CWUser'

        class BaseRewriter(SchemaBasedRewriter):
            rules = [
               (rgx('/collector(.*)'),
                rgx_action(rql=RQL1,
                    form=dict(vid='baseindex')),
                ),
                ]
        class Rewriter(BaseRewriter):
            rules = [
               (rgx('/collector(.*)'),
                rgx_action(rql=RQL2,
                    form=dict(vid='index')),
                ),
                ]

        with self.admin_access.web_request() as req:
            rewriter = Rewriter(req)
            _pmid, rset = rewriter.rewrite(req, '/collector')
            self.assertEqual(rset.rql, RQL2)
            self.assertEqual(req.form, {'vid' : "index"})
            _pmid, rset = rewriter.rewrite(req, '/collector/something')
            self.assertEqual(rset.rql, RQL2)
            self.assertEqual(req.form, {'vid' : "index"})
            _pmid, rset = rewriter.rewrite(req, '/collector/something/')
            self.assertEqual(req.form, {'vid' : "index"})
            self.assertEqual(rset.rql, RQL2)
            _pmid, rset = rewriter.rewrite(req, '/collector/somethingelse/')
            self.assertEqual(rset.rql, RQL2)
            self.assertEqual(req.form, {'vid' : "index"})


if __name__ == '__main__':
    from logilab.common.testlib import unittest_main
    unittest_main()



# copyright 2003-2012 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of CubicWeb.
#
# CubicWeb is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# CubicWeb is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with CubicWeb.  If not, see <http://www.gnu.org/licenses/>.
"""Specific views for CWProperty (eg site/user preferences"""


from cubicweb import _

from logilab.mtconverter import xml_escape

from logilab.common.decorators import cached

from cubicweb import UnknownProperty
from cubicweb.predicates import (one_line_rset, none_rset, is_instance,
                                 match_user_groups, logged_user_in_rset)
from cubicweb.view import StartupView
from cubicweb.web import stdmsgs
from cubicweb.web.form import FormViewMixIn
from cubicweb.web.formfields import FIELDS, StringField
from cubicweb.web.formwidgets import (Select, TextInput, Button, SubmitButton,
                                      FieldWidget)
from cubicweb.web.views import uicfg, primary, formrenderers, editcontroller
from cubicweb.web.views.ajaxcontroller import ajaxfunc

uicfg.primaryview_section.tag_object_of(('*', 'for_user', '*'), 'hidden')

# some string we want to be internationalizable for nicer display of property
# groups
_('navigation')
_('ui')
_('boxes')
_('components')
_('ctxcomponents')
_('navigation.combobox-limit')
_('navigation.page-size')
_('navigation.related-limit')
_('navigation.short-line-size')
_('ui.date-format')
_('ui.datetime-format')
_('ui.default-text-format')
_('ui.fckeditor')
_('ui.float-format')
_('ui.language')
_('ui.time-format')
_('open all')
_('ui.main-template')
_('ui.site-title')
_('ui.encoding')
_('category')


def make_togglable_link(nodeid, label):
    """builds a HTML link that switches the visibility & remembers it"""
    return u'<a href="javascript: togglePrefVisibility(\'%s\')">%s</a>' % (
        nodeid, label)

def css_class(someclass):
    return someclass and 'class="%s"' % someclass or ''


class CWPropertyPrimaryView(primary.PrimaryView):
    __select__ = is_instance('CWProperty')
    skip_none = False


class SystemCWPropertiesForm(FormViewMixIn, StartupView):
    """site-wide properties edition form"""
    __regid__ = 'systempropertiesform'
    __select__ = none_rset() & match_user_groups('managers')
    form_buttons = [SubmitButton()]

    title = _('site configuration')
    category = 'startupview'

    def linkable(self):
        return True

    def url(self):
        """return the url associated with this view. We can omit rql here"""
        return self._cw.build_url('view', vid=self.__regid__)

    def _cookie_name(self, somestr):
        return str('%s_property_%s' % (self._cw.vreg.config.appid, somestr))

    def _group_status(self, group, default=u'hidden'):
        """return css class name 'hidden' (collapsed), or '' (open)"""
        cookies = self._cw.get_cookie()
        cookiename = self._cookie_name(group)
        cookie = cookies.get(cookiename)
        if cookie is None:
            self._cw.set_cookie(cookiename, default, maxage=None)
            status = default
        else:
            status = cookie.value
        return status

    def call(self, **kwargs):
        self._cw.add_js(('cubicweb.preferences.js',
                         'cubicweb.edition.js', 'cubicweb.ajax.js'))
        self._cw.add_css('cubicweb.preferences.css')
        values = self.defined_keys
        mainopts, groupedopts = self.group_properties()
        # precompute all forms first to consume error message
        mainforms, groupedforms = self.build_forms(mainopts, groupedopts)
        _ = self._cw._
        self.w(u'<h1>%s</h1>\n' % _(self.title))
        for label, group, form in sorted((_(g), g, f)
                                         for g, f in mainforms.items()):
            self.wrap_main_form(group, label, form)
        for label, group, objects in sorted((_(g), g, o)
                                            for g, o in groupedforms.items()):
            self.wrap_grouped_form(group, label, objects)

    @property
    @cached
    def cwprops_rset(self):
        return self._cw.execute('Any P,K,V WHERE P is CWProperty, P pkey K, '
                                'P value V, NOT P for_user U')

    @property
    def defined_keys(self):
        values = {}
        for i, entity in enumerate(self.cwprops_rset.entities()):
            values[entity.pkey] = i
        return values

    def group_properties(self):
        mainopts, groupedopts = {}, {}
        vreg = self._cw.vreg
        # "self._regid__=='systempropertiesform'" to skip site wide properties on
        # user's preference but not site's configuration
        for key in vreg.user_property_keys(self.__regid__=='systempropertiesform'):
            parts = key.split('.')
            if parts[0] in vreg and len(parts) >= 3:
                # appobject configuration
                reg = parts[0]
                propid = parts[-1]
                oid = '.'.join(parts[1:-1])
                groupedopts.setdefault(reg, {}).setdefault(oid, []).append(key)
            else:
                mainopts.setdefault(parts[0], []).append(key)
        return mainopts, groupedopts

    def build_forms(self, mainopts, groupedopts):
        mainforms, groupedforms = {}, {}
        for group, keys in mainopts.items():
            mainforms[group] = self.form(group, keys, False)
        for group, objects in groupedopts.items():
            groupedforms[group] = {}
            for oid, keys in objects.items():
                groupedforms[group][oid] = self.form(group + '_' + oid, keys, True)
        return mainforms, groupedforms

    def entity_for_key(self, key):
        values = self.defined_keys
        if key in values:
            entity = self.cwprops_rset.get_entity(values[key], 0)
        else:
            entity = self._cw.vreg['etypes'].etype_class('CWProperty')(self._cw)
            entity.eid = next(self._cw.varmaker)
            entity.cw_attr_cache['pkey'] = key
            entity.cw_attr_cache['value'] = self._cw.vreg.property_value(key)
        return entity

    def form(self, formid, keys, splitlabel=False):
        form = self._cw.vreg['forms'].select(
            'composite', self._cw, domid=formid, action=self._cw.build_url(),
            form_buttons=self.form_buttons,
            onsubmit="return validatePrefsForm('%s')" % formid,
            submitmsg=self._cw._('changes applied'))
        path = self._cw.relative_path()
        if '?' in path:
            path, params = path.split('?', 1)
            form.add_hidden('__redirectparams', params)
        form.add_hidden('__redirectpath', path)
        for key in keys:
            self.form_row(form, key, splitlabel)
        renderer = self._cw.vreg['formrenderers'].select('cwproperties', self._cw,
                                                     display_progress_div=False)
        data = []
        form.render(w=data.append, renderer=renderer)
        return u'\n'.join(data)

    def form_row(self, form, key, splitlabel):
        entity = self.entity_for_key(key)
        if splitlabel:
            label = key.split('.')[-1]
        else:
            label = key
        subform = self._cw.vreg['forms'].select('base', self._cw, entity=entity,
                                                mainform=False)
        subform.append_field(PropertyValueField(name='value', label=label, role='subject',
                                                eidparam=True))
        subform.add_hidden('pkey', key, eidparam=True, role='subject')
        form.add_subform(subform)
        return subform

    def wrap_main_form(self, group, label, form):
        status = css_class(self._group_status(group))
        self.w(u'<div class="propertiesform">%s</div>\n' %
               (make_togglable_link('fieldset_' + group, label)))
        self.w(u'<div id="fieldset_%s" %s>' % (group, status))
        self.w(u'<fieldset class="preferences">')
        self.w(form)
        self.w(u'</fieldset></div>')

    def wrap_grouped_form(self, group, label, objects):
        status = css_class(self._group_status(group))
        self.w(u'<div class="propertiesform">%s</div>\n' %
          (make_togglable_link('fieldset_' + group, label)))
        self.w(u'<div id="fieldset_%s" %s>' % (group, status))
        sorted_objects = sorted((self._cw.__('%s_%s' % (group, o)), o, f)
                                for o, f in objects.items())
        for label, oid, form in sorted_objects:
            self.wrap_object_form(group, oid, label, form)
        self.w(u'</div>')

    def wrap_object_form(self, group, oid, label, form):
        w = self.w
        w(u'<div class="component">')
        w(u'''<div class="componentLink"><a href="javascript:$.noop();"
                   onclick="javascript:toggleVisibility('field_%(oid)s_%(group)s')"
                   class="componentTitle">%(label)s</a>''' % {'label':label, 'oid':oid, 'group':group})
        w(u''' (<div class="openlink"><a href="javascript:$.noop();"
                onclick="javascript:openFieldset('fieldset_%(group)s')">%(label)s</a></div>)'''
                  % {'label':self._cw._('open all'), 'group':group})
        w(u'</div>')
        docmsgid = '%s_%s_description' % (group, oid)
        doc = self._cw._(docmsgid)
        if doc != docmsgid:
            w(u'<div class="helper">%s</div>' % xml_escape(doc).capitalize())
        w(u'</div>')
        w(u'<fieldset id="field_%(oid)s_%(group)s" class="%(group)s preferences hidden">'
          % {'oid':oid, 'group':group})
        w(form)
        w(u'</fieldset>')


class CWPropertiesForm(SystemCWPropertiesForm):
    """user's preferences properties edition form"""
    __regid__ = 'propertiesform'
    __select__ = (
        (none_rset() & match_user_groups('users','managers'))
        | (one_line_rset() & match_user_groups('users') & logged_user_in_rset())
        | (one_line_rset() & match_user_groups('managers') & is_instance('CWUser'))
        )

    title = _('user preferences')

    @property
    def user(self):
        if self.cw_rset is None:
            return self._cw.user
        return self.cw_rset.get_entity(self.cw_row or 0, self.cw_col or 0)

    @property
    @cached
    def cwprops_rset(self):
        return self._cw.execute('Any P,K,V WHERE P is CWProperty, P pkey K, P value V,'
                                'P for_user U, U eid %(x)s', {'x': self.user.eid})

    def form_row(self, form, key, splitlabel):
        subform = super(CWPropertiesForm, self).form_row(form, key, splitlabel)
        # if user is in the managers group and the property is being created,
        # we have to set for_user explicitly
        if not subform.edited_entity.has_eid() and self.user.matching_groups('managers'):
            subform.add_hidden('for_user', self.user.eid, eidparam=True, role='subject')
        return subform

# cwproperty form objects ######################################################

class PlaceHolderWidget(FieldWidget):

    def render(self, form, field, renderer):
        domid = field.dom_id(form)
        # empty span as well else html validation fail (label is refering to
        # this id)
        return '<div id="div:%s"><span id="%s">%s</span></div>' % (
            domid, domid, form._cw._('select a key first'))


class NotEditableWidget(FieldWidget):
    def __init__(self, value, msg=None):
        self.value = value
        self.msg = msg

    def render(self, form, field, renderer):
        domid = field.dom_id(form)
        value = '<span class="value" id="%s">%s</span>' % (domid, self.value)
        if self.msg:
            value += '<div class="helper">%s</div>' % self.msg
        return value


class PropertyKeyField(StringField):
    """specific field for CWProperty.pkey to set the value widget according to
    the selected key
    """
    widget = Select

    def render(self, form, renderer):
        wdg = self.get_widget(form)
        # pylint: disable=E1101
        wdg.attrs['onchange'] = "javascript:setPropValueWidget('%s')" % (
            form.edited_entity.eid)
        return wdg.render(form, self, renderer)

    def vocabulary(self, form):
        entity = form.edited_entity
        _ = form._cw._
        if entity.has_eid():
            return [(_(entity.pkey), entity.pkey)]
        choices = entity._cw.vreg.user_property_keys()
        return [(u'', u'')] + sorted(zip((_(v) for v in choices), choices))


class PropertyValueField(StringField):
    """specific field for CWProperty.value  which will be different according to
    the selected key type and vocabulary information
    """
    widget = PlaceHolderWidget

    def render(self, form, renderer=None):
        wdg = self.get_widget(form)
        return wdg.render(form, self, renderer)

    def form_init(self, form):
        entity = form.edited_entity
        if not (entity.has_eid() or 'pkey' in entity.cw_attr_cache):
            # no key set yet, just include an empty div which will be filled
            # on key selection
            return
        try:
            pdef = form._cw.vreg.property_info(entity.pkey)
        except UnknownProperty as ex:
            form.warning('%s (you should probably delete that property '
                         'from the database)', ex)
            msg = form._cw._('you should probably delete that property')
            self.widget = NotEditableWidget(entity.printable_value('value'),
                                            '%s (%s)' % (msg, ex))
            return
        if entity.pkey.startswith('system.'):
            msg = form._cw._('value associated to this key is not editable '
                             'manually')
            self.widget = NotEditableWidget(entity.printable_value('value'), msg)
        # XXX race condition when used from CWPropertyForm, should not rely on
        # instance attributes
        self.value = pdef['default']
        self.help = pdef['help']
        vocab = pdef['vocabulary']
        if vocab is not None:
            if callable(vocab):
                # list() just in case its a generator function
                self.choices = list(vocab())
            else:
                self.choices = vocab
            wdg = Select()
        elif pdef['type'] == 'String': # else we'll get a TextArea by default
            wdg = TextInput()
        else:
            field = FIELDS[pdef['type']]()
            wdg = field.widget
            if pdef['type'] == 'Boolean':
                self.choices = field.vocabulary(form)
        self.widget = wdg


class CWPropertiesFormRenderer(formrenderers.FormRenderer):
    """specific renderer for properties"""
    __regid__ = 'cwproperties'

    def open_form(self, form, values):
        err = '<div class="formsg"></div>'
        return super(CWPropertiesFormRenderer, self).open_form(form, values) + err

    def _render_fields(self, fields, w, form):
        for field in fields:
            w(u'<div class="preffield">\n')
            if self.display_label:
                w(u'%s' % self.render_label(form, field))
            error = form.field_error(field)
            if error:
                w(u'<span class="error">%s</span>' % error)
            w(u'%s' % self.render_help(form, field))
            w(u'<div class="prefinput">')
            w(field.render(form, self))
            w(u'</div>')
            w(u'</div>')

    def render_buttons(self, w, form):
        w(u'<div>\n')
        for button in form.form_buttons:
            w(u'%s\n' % button.render(form))
        w(u'</div>')


class CWPropertyIEditControlAdapter(editcontroller.IEditControlAdapter):
    __select__ = is_instance('CWProperty')

    def after_deletion_path(self):
        """return (path, parameters) which should be used as redirect
        information when this entity is being deleted
        """
        return 'view', {}


@ajaxfunc(output_type='xhtml')
def prop_widget(self, propkey, varname):
    """specific method for CWProperty handling"""
    entity = self._cw.vreg['etypes'].etype_class('CWProperty')(self._cw)
    entity.eid = varname
    entity.pkey = propkey
    form = self._cw.vreg['forms'].select('edition', self._cw, entity=entity)
    form.build_context()
    vfield = form.field_by_name('value', 'subject')
    renderer = formrenderers.FormRenderer(self._cw)
    return vfield.render(form, renderer) \
           + renderer.render_help(form, vfield)

_afs = uicfg.autoform_section
_afs.tag_subject_of(('*', 'for_user', '*'), 'main', 'hidden')
_afs.tag_object_of(('*', 'for_user', '*'), 'main', 'hidden')
_aff = uicfg.autoform_field
_aff.tag_attribute(('CWProperty', 'pkey'), PropertyKeyField)
_aff.tag_attribute(('CWProperty', 'value'), PropertyValueField)



# copyright 2003-2011 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of CubicWeb.
#
# CubicWeb is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# CubicWeb is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with CubicWeb.  If not, see <http://www.gnu.org/licenses/>.
"""utilities for instances migration"""
from __future__ import print_function



import sys
import os
import logging
import tempfile
from os.path import exists, join, basename, splitext
from itertools import chain
from warnings import warn

from six import string_types

from logilab.common import IGNORED_EXTENSIONS
from logilab.common.decorators import cached
from logilab.common.configuration import REQUIRED, read_old_config
from logilab.common.shellutils import ASK
from logilab.common.changelog import Version
from logilab.common.deprecation import deprecated

from cubicweb import ConfigurationError, ExecutionError
from cubicweb.cwconfig import CubicWebConfiguration as cwcfg
from cubicweb.toolsutils import show_diffs

def filter_scripts(config, directory, fromversion, toversion, quiet=True):
    """return a list of paths of migration files to consider to upgrade
    from a version to a greater one
    """
    from logilab.common.changelog import Version # doesn't work with appengine
    assert fromversion
    assert toversion
    assert isinstance(fromversion, tuple), fromversion.__class__
    assert isinstance(toversion, tuple), toversion.__class__
    assert fromversion <= toversion, (fromversion, toversion)
    if not exists(directory):
        if not quiet:
            print(directory, "doesn't exists, no migration path")
        return []
    if fromversion == toversion:
        return []
    result = []
    for fname in os.listdir(directory):
        if fname.endswith(IGNORED_EXTENSIONS):
            continue
        fpath = join(directory, fname)
        try:
            tver, mode = fname.split('_', 1)
        except ValueError:
            continue
        mode = mode.split('.', 1)[0]
        if not config.accept_mode(mode):
            continue
        try:
            tver = Version(tver)
        except ValueError:
            continue
        if tver <= fromversion:
            continue
        if tver > toversion:
            continue
        result.append((tver, fpath))
    # be sure scripts are executed in order
    return sorted(result)


def execscript_confirm(scriptpath):
    """asks for confirmation before executing a script and provides the
    ability to show the script's content
    """
    while True:
        answer = ASK.ask('Execute %r ?' % scriptpath,
                         ('Y','n','show','abort'), 'Y')
        if answer == 'abort':
            raise SystemExit(1)
        elif answer == 'n':
            return False
        elif answer == 'show':
            stream = open(scriptpath)
            scriptcontent = stream.read()
            stream.close()
            print()
            print(scriptcontent)
            print()
        else:
            return True

def yes(*args, **kwargs):
    return True


class MigrationHelper(object):
    """class holding CubicWeb Migration Actions used by migration scripts"""

    def __init__(self, config, interactive=True, verbosity=1):
        self.config = config
        if config:
            # no config on shell to a remote instance
            self.config.init_log(logthreshold=logging.ERROR)
        # 0: no confirmation, 1: only main commands confirmed, 2 ask for everything
        self.verbosity = verbosity
        self.need_wrap = True
        if not interactive or not verbosity:
            self.confirm = yes
            self.execscript_confirm = yes
        else:
            self.execscript_confirm = execscript_confirm
        self._option_changes = []
        self.__context = {'confirm': self.confirm,
                          'config': self.config,
                          'interactive_mode': interactive,
                          }
        self._context_stack = []

    def __getattribute__(self, name):
        try:
            return object.__getattribute__(self, name)
        except AttributeError:
            cmd = 'cmd_%s' % name
            # search self.__class__ to avoid infinite recursion
            if hasattr(self.__class__, cmd):
                meth = getattr(self, cmd)
                return lambda *args, **kwargs: self.interact(args, kwargs,
                                                             meth=meth)
            raise
        raise AttributeError(name)

    def migrate(self, vcconf, toupgrade, options):
        """upgrade the given set of cubes

        `cubes` is an ordered list of 3-uple:
        (cube, fromversion, toversion)
        """
        if options.fs_only:
            # monkey path configuration.accept_mode so database mode (e.g. Any)
            # won't be accepted
            orig_accept_mode = self.config.accept_mode
            def accept_mode(mode):
                if mode == 'Any':
                    return False
                return orig_accept_mode(mode)
            self.config.accept_mode = accept_mode
        # may be an iterator
        toupgrade = tuple(toupgrade)
        vmap = dict( (cube, (fromver, tover)) for cube, fromver, tover in toupgrade)
        ctx = self.__context
        ctx['versions_map'] = vmap
        if self.config.accept_mode('Any') and 'cubicweb' in vmap:
            migrdir = self.config.migration_scripts_dir()
            self.cmd_process_script(join(migrdir, 'bootstrapmigration_repository.py'))
        for cube, fromversion, toversion in toupgrade:
            if cube == 'cubicweb':
                migrdir = self.config.migration_scripts_dir()
            else:
                migrdir = self.config.cube_migration_scripts_dir(cube)
            scripts = filter_scripts(self.config, migrdir, fromversion, toversion)
            if scripts:
                prevversion = None
                for version, script in scripts:
                    # take care to X.Y.Z_Any.py / X.Y.Z_common.py: we've to call
                    # cube_upgraded once all script of X.Y.Z have been executed
                    if prevversion is not None and version != prevversion:
                        self.cube_upgraded(cube, prevversion)
                    prevversion = version
                    self.cmd_process_script(script)
                self.cube_upgraded(cube, toversion)
            else:
                self.cube_upgraded(cube, toversion)

    def cube_upgraded(self, cube, version):
        pass

    def shutdown(self):
        pass

    def interact(self, args, kwargs, meth):
        """execute the given method according to user's confirmation"""
        msg = 'Execute command: %s(%s) ?' % (
            meth.__name__[4:],
            ', '.join([repr(arg) for arg in args] +
                      ['%s=%r' % (n,v) for n,v in kwargs.items()]))
        if 'ask_confirm' in kwargs:
            ask_confirm = kwargs.pop('ask_confirm')
        else:
            ask_confirm = True
        if not ask_confirm or self.confirm(msg):
            return meth(*args, **kwargs)

    def confirm(self, question, # pylint: disable=E0202
                shell=True, abort=True, retry=False, pdb=False, default='y'):
        """ask for confirmation and return true on positive answer

        if `retry` is true the r[etry] answer may return 2
        """
        possibleanswers = ['y', 'n']
        if abort:
            possibleanswers.append('abort')
        if pdb:
            possibleanswers.append('pdb')
        if shell:
            possibleanswers.append('shell')
        if retry:
            possibleanswers.append('retry')
        try:
            answer = ASK.ask(question, possibleanswers, default)
        except (EOFError, KeyboardInterrupt):
            answer = 'abort'
        if answer == 'n':
            return False
        if answer == 'retry':
            return 2
        if answer == 'abort':
            raise SystemExit(1)
        if answer == 'shell':
            self.interactive_shell()
            return self.confirm(question, shell, abort, retry, pdb, default)
        if answer == 'pdb':
            import pdb
            pdb.set_trace()
            return self.confirm(question, shell, abort, retry, pdb, default)
        return True

    def interactive_shell(self):
        self.confirm = yes
        self.need_wrap = False
        # avoid '_' to be added to builtins by sys.display_hook
        def do_not_add___to_builtins(obj):
            if obj is not None:
                print(repr(obj))
        sys.displayhook = do_not_add___to_builtins
        local_ctx = self._create_context()
        try:
            import readline
            from cubicweb.toolsutils import CWShellCompleter
        except ImportError:
            # readline not available
            pass
        else:
            rql_completer = CWShellCompleter(local_ctx)
            readline.set_completer(rql_completer.complete)
            readline.parse_and_bind('tab: complete')
            home_key = 'HOME'
            if sys.platform == 'win32':
                home_key = 'USERPROFILE'
            histfile = os.path.join(os.environ[home_key], ".cwshell_history")
            try:
                readline.read_history_file(histfile)
            except IOError:
                pass
        from code import interact
        banner = """entering the migration python shell
just type migration commands or arbitrary python code and type ENTER to execute it
type "exit" or Ctrl-D to quit the shell and resume operation"""
        interact(banner, local=local_ctx)
        try:
            readline.write_history_file(histfile)
        except IOError:
            pass
        # delete instance's confirm attribute to avoid questions
        del self.confirm
        self.need_wrap = True

    @cached
    def _create_context(self):
        """return a dictionary to use as migration script execution context"""
        context = self.__context
        for attr in dir(self):
            if attr.startswith('cmd_'):
                if self.need_wrap:
                    context[attr[4:]] = getattr(self, attr[4:])
                else:
                    context[attr[4:]] = getattr(self, attr)
        return context

    def update_context(self, key, value):
        for context in self._context_stack:
            context[key] = value
        self.__context[key] = value

    def cmd_process_script(self, migrscript, funcname=None, *args, **kwargs):
        """execute a migration script in interactive mode

        Display the migration script path, ask for confirmation and execute it
        if confirmed

        Allowed input file formats for migration scripts:
        - `python` (.py)
        - `sql` (.sql)
        - `doctest` (.txt or .rst)

        .. warning:: sql migration scripts are not available in web-only instance

        You can pass script parameters with using double dash (--) in the
        command line

        Context environment can have these variables defined:
        - __name__ : will be determine by funcname parameter
        - __file__ : is the name of the script if it exists
        - __args__ : script arguments coming from command-line

        :param migrscript: name of the script
        :param funcname: defines __name__ inside the shell (or use __main__)
        :params args: optional arguments for funcname
        :keyword scriptargs: optional arguments of the script
        """
        ftypes = {'python':  ('.py',),
                  'doctest': ('.txt', '.rst'),
                  'sql':     ('.sql',)}
        # sql migration scripts are not available in web-only instance
        if not hasattr(self, "session"):
            ftypes.pop('sql')
        migrscript = os.path.normpath(migrscript)
        for (script_mode, ftype) in ftypes.items():
            if migrscript.endswith(ftype):
                break
        else:
            ftypes = ', '.join(chain(*ftypes.values()))
            msg = 'ignoring %s, not a valid script extension (%s)'
            raise ExecutionError(msg % (migrscript, ftypes))
        if not self.execscript_confirm(migrscript):
            return
        scriptlocals = self._create_context().copy()
        scriptlocals.update({'__file__': migrscript,
                             '__args__': kwargs.pop("scriptargs", [])})
        self._context_stack.append(scriptlocals)
        if script_mode == 'python':
            if funcname is None:
                pyname = '__main__'
            else:
                pyname = splitext(basename(migrscript))[0]
            scriptlocals['__name__'] = pyname
            with open(migrscript, 'rb') as fobj:
                fcontent = fobj.read()
            try:
                code = compile(fcontent, migrscript, 'exec')
            except SyntaxError:
                # try without print_function
                code = compile(fcontent, migrscript, 'exec', 0, True)
                warn('[3.22] script %r should be updated to work with print_function'
                     % migrscript, DeprecationWarning)
            exec(code, scriptlocals)
            if funcname is not None:
                try:
                    func = scriptlocals[funcname]
                    self.info('found %s in locals', funcname)
                    assert callable(func), '%s (%s) is not callable' % (func, funcname)
                except KeyError:
                    self.critical('no %s in script %s', funcname, migrscript)
                    return None
                return func(*args, **kwargs)
        elif script_mode == 'sql':
            from cubicweb.server.sqlutils import sqlexec
            sqlexec(open(migrscript).read(), self.session.system_sql)
            self.commit()
        else: # script_mode == 'doctest'
            import doctest
            return doctest.testfile(migrscript, module_relative=False,
                                    optionflags=doctest.ELLIPSIS,
                                    # verbose mode when user input is expected
                                    verbose=self.verbosity==2,
                                    report=True,
                                    encoding='utf-8',
                                    globs=scriptlocals)
        self._context_stack.pop()

    def cmd_option_renamed(self, oldname, newname):
        """a configuration option has been renamed"""
        self._option_changes.append(('renamed', oldname, newname))

    def cmd_option_group_changed(self, option, oldgroup, newgroup):
        """a configuration option has been moved in another group"""
        self._option_changes.append(('moved', option, oldgroup, newgroup))

    def cmd_option_added(self, optname):
        """a configuration option has been added"""
        self._option_changes.append(('added', optname))

    def cmd_option_removed(self, optname):
        """a configuration option has been removed"""
        # can safely be ignored
        #self._option_changes.append(('removed', optname))

    def cmd_option_type_changed(self, optname, oldtype, newvalue):
        """a configuration option's type has changed"""
        self._option_changes.append(('typechanged', optname, oldtype, newvalue))

    def cmd_add_cubes(self, cubes):
        """modify the list of used cubes in the in-memory config
        returns newly inserted cubes, including dependencies
        """
        if isinstance(cubes, string_types):
            cubes = (cubes,)
        origcubes = self.config.cubes()
        newcubes = [p for p in self.config.expand_cubes(cubes)
                    if not p in origcubes]
        if newcubes:
            self.config.add_cubes(newcubes)
        return newcubes

    @deprecated('[3.20] use drop_cube() instead of remove_cube()')
    def cmd_remove_cube(self, cube, removedeps=False):
        return self.cmd_drop_cube(cube, removedeps)

    def cmd_drop_cube(self, cube, removedeps=False):
        if removedeps:
            toremove = self.config.expand_cubes([cube])
        else:
            toremove = (cube,)
        origcubes = self.config._cubes
        basecubes = [c for c in origcubes if not c in toremove]
        # don't fake-add any new ones, or we won't be able to really-add them later
        self.config._cubes = tuple(cube for cube in self.config.expand_cubes(basecubes)
                                   if cube in origcubes)
        removed = [p for p in origcubes if not p in self.config._cubes]
        if not cube in removed and cube in origcubes:
            raise ConfigurationError("can't remove cube %s, "
                                     "used as a dependency" % cube)
        return removed

    def rewrite_configuration(self):
        configfile = self.config.main_config_file()
        if self._option_changes:
            read_old_config(self.config, self._option_changes, configfile)
        fd, newconfig = tempfile.mkstemp()
        for optdescr in self._option_changes:
            if optdescr[0] == 'added':
                optdict = self.config.get_option_def(optdescr[1])
                if optdict.get('default') is REQUIRED:
                    self.config.input_option(optdescr[1], optdict)
        self.config.generate_config(open(newconfig, 'w'))
        show_diffs(configfile, newconfig, askconfirm=self.confirm is not yes)
        os.close(fd)
        if exists(newconfig):
            os.unlink(newconfig)

    # these are overridden by set_log_methods below
    # only defining here to prevent pylint from complaining
    info = warning = error = critical = exception = debug = lambda msg,*a,**kw: None

from logging import getLogger
from cubicweb import set_log_methods
set_log_methods(MigrationHelper, getLogger('cubicweb.migration'))


def version_strictly_lower(a, b):
    if a is None:
        return True
    if b is None:
        return False
    if a:
        a = Version(a)
    if b:
        b = Version(b)
    return a < b

def max_version(a, b):
    return str(max(Version(a), Version(b)))

class ConfigurationProblem(object):
    """Each cube has its own list of dependencies on other cubes/versions.

    The ConfigurationProblem is used to record the loaded cubes, then to detect
    inconsistencies in their dependencies.

    See configuration management on wikipedia for litterature.
    """

    def __init__(self, config):
        self.config = config
        self.cubes = {'cubicweb': cwcfg.cubicweb_version()}

    def add_cube(self, name, version):
        self.cubes[name] = version

    def solve(self):
        self.warnings = []
        self.errors = []
        self.dependencies = {}
        self.reverse_dependencies = {}
        self.constraints = {}
        # read dependencies
        for cube in self.cubes:
            if cube == 'cubicweb': continue
            self.dependencies[cube] = dict(self.config.cube_dependencies(cube))
            self.dependencies[cube]['cubicweb'] = self.config.cube_depends_cubicweb_version(cube)
        # compute reverse dependencies
        for cube, dependencies in self.dependencies.items():
            for name, constraint in dependencies.items():
                self.reverse_dependencies.setdefault(name,set())
                if constraint:
                    try:
                        oper, version = constraint.split()
                        self.reverse_dependencies[name].add( (oper, version, cube) )
                    except Exception:
                        self.warnings.append(
                            'cube %s depends on %s but constraint badly '
                            'formatted: %s' % (cube, name, constraint))
                else:
                    self.reverse_dependencies[name].add( (None, None, cube) )
        # check consistency
        for cube, versions in sorted(self.reverse_dependencies.items()):
            oper, version, source = None, None, None
            # simplify constraints
            if versions:
                for constraint in versions:
                    op, ver, src = constraint
                    if oper is None:
                        oper = op
                        version = ver
                        source = src
                    elif op == '>=' and oper == '>=':
                        if version_strictly_lower(version, ver):
                            version = ver
                            source = src
                    elif op == None:
                        continue
                    else:
                        print('unable to handle %s in %s, set to `%s %s` '
                              'but currently up to `%s %s`' %
                              (cube, source, oper, version, op, ver))
            # "solve" constraint satisfaction problem
            if cube not in self.cubes:
                self.errors.append( ('add', cube, version, source) )
            elif versions:
                lower_strict = version_strictly_lower(self.cubes[cube], version)
                if oper in ('>=','=','=='):
                    if lower_strict:
                        self.errors.append( ('update', cube, version, source) )
                elif oper is None:
                    pass # no constraint on version
                else:
                    print('unknown operator', oper)



# copyright 2003-2012 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of CubicWeb.
#
# CubicWeb is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# CubicWeb is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with CubicWeb.  If not, see <http://www.gnu.org/licenses/>.
"""

The `AppObject` class
---------------------

The AppObject class is the base class for all dynamically loaded objects
(application objects) accessible through the vregistry.

We can find a certain number of attributes and methods defined in this class and
common to all the application objects.

"""


from logging import getLogger

from logilab.common.deprecation import deprecated, class_renamed
from logilab.common.logging_ext import set_log_methods

# first line imports for bw compat
from logilab.common.registry import (objectify_predicate, traced_selection, Predicate,
                                     RegistrableObject, yes)


objectify_selector = deprecated('[3.15] objectify_selector has been '
                                'renamed to objectify_predicates in '
                                'logilab.common.registry')(objectify_predicate)
traced_selection = deprecated('[3.15] traced_selection has been '
                              'moved to logilab.common.registry')(traced_selection)
Selector = class_renamed('Selector', Predicate,
                         '[3.15] Selector has been renamed to Predicate '
                         'in logilab.common.registry')

@deprecated('[3.15] lltrace decorator can now be removed')
def lltrace(func):
    return func

# the base class for all appobjects ############################################

class AppObject(RegistrableObject):
    """This is the base class for CubicWeb application objects which are
    selected in a request context.

    The following attributes should be set on concrete appobject classes:

    At selection time, the following attributes are set on the instance:

    :attr:`_cw`
      current request
    :attr:`cw_extra_kwargs`
      other received arguments

    And also the following, only if `rset` is found in arguments (in which case
    rset/row/col will be removed from `cwextra_kwargs`):

    :attr:`cw_rset`
      context result set or None

    :attr:`cw_row`
      if a result set is set and the context is about a particular cell in the
      result set, and not the result set as a whole, specify the row number we
      are interested in, else None

    :attr:`cw_col`
      if a result set is set and the context is about a particular cell in the
      result set, and not the result set as a whole, specify the col number we
      are interested in, else None


    .. Note::

      * do not inherit directly from this class but from a more specific class
        such as `AnyEntity`, `EntityView`, `AnyRsetView`, `Action`...

    """
    __select__ = yes()

    @classmethod
    def __registered__(cls, registry):
        """called by the registry when the appobject has been registered.

        It must return the object that will be actually registered (this may be
        the right hook to create an instance for example). By default the
        appobject is returned without any transformation.
        """
        pdefs = getattr(cls, 'cw_property_defs', {})
        for propid, pdef in pdefs.items():
            pdef = pdef.copy() # may be shared
            pdef['default'] = getattr(cls, propid, pdef['default'])
            pdef['sitewide'] = getattr(cls, 'site_wide', pdef.get('sitewide'))
            registry.vreg.register_property(cls._cwpropkey(propid), **pdef)
        assert callable(cls.__select__), cls
        return cls

    def __init__(self, req, **extra):
        super(AppObject, self).__init__()
        self._cw = req
        try:
            self.cw_rset = extra.pop('rset')
            self.cw_row = extra.pop('row', None)
            self.cw_col = extra.pop('col', None)
        except KeyError:
            pass
        self.cw_extra_kwargs = extra

    # persistent class properties ##############################################
    #
    # optional `cw_property_defs` dict on a class defines available persistent
    # properties for this class:
    #
    # * key: id of the property (the actual CWProperty key is build using
    #        <registry name>.<obj id>.<property id>
    # * value: tuple (property type, vocabfunc, default value, property description)
    #         possible types are those used by `logilab.common.configuration`
    #
    # notice that when it exists multiple objects with the same id (adaptation,
    # overriding) only the first encountered definition is considered, so those
    # objects can't try to have different default values for instance.
    #
    # you can then access to a property value using self.cw_propval, where self
    # is an instance of class

    @classmethod
    def _cwpropkey(cls, propid):
        """return cw property key for the property of the given id for this
        class
        """
        return '%s.%s.%s' % (cls.__registry__, cls.__regid__, propid)

    def cw_propval(self, propid):
        """return cw property value associated to key

        <cls.__registry__>.<cls.id>.<propid>
        """
        return self._cw.property_value(self._cwpropkey(propid))

    # these are overridden by set_log_methods below
    # only defining here to prevent pylint from complaining
    info = warning = error = critical = exception = debug = lambda msg,*a,**kw: None

set_log_methods(AppObject, getLogger('cubicweb.appobject'))

# defined here to avoid warning on usage on the AppObject class
yes = deprecated('[3.15] yes has been moved to logilab.common.registry')(yes)



# copyright 2010-2016 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of CubicWeb.
#
# CubicWeb is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# CubicWeb is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with CubicWeb.  If not, see <http://www.gnu.org/licenses/>.
"""hooks for repository sources synchronization"""

from cubicweb import _

from socket import gethostname

from logilab.common.decorators import clear_cache

from cubicweb import validation_error
from cubicweb.predicates import is_instance
from cubicweb.server import SOURCE_TYPES, hook


class SourceHook(hook.Hook):
    __abstract__ = True
    category = 'cw.sources'


# repo sources synchronization #################################################

class SourceAddedOp(hook.Operation):
    entity = None # make pylint happy
    def postcommit_event(self):
        self.cnx.repo.add_source(self.entity)

class SourceAddedHook(SourceHook):
    __regid__ = 'cw.sources.added'
    __select__ = SourceHook.__select__ & is_instance('CWSource')
    events = ('after_add_entity',)
    def __call__(self):
        try:
            sourcecls = SOURCE_TYPES[self.entity.type]
        except KeyError:
            msg = _('Unknown source type')
            raise validation_error(self.entity, {('type', 'subject'): msg})
        # ignore creation of the system source done during database
        # initialisation, as config for this source is in a file and handling
        # is done separatly (no need for the operation either)
        if self.entity.name != 'system':
            sourcecls.check_conf_dict(self.entity.eid, self.entity.host_config,
                                      fail_if_unknown=not self._cw.vreg.config.repairing)
            SourceAddedOp(self._cw, entity=self.entity)


class SourceRemovedOp(hook.Operation):
    uri = None # make pylint happy
    def postcommit_event(self):
        self.cnx.repo.remove_source(self.uri)

class SourceRemovedHook(SourceHook):
    __regid__ = 'cw.sources.removed'
    __select__ = SourceHook.__select__ & is_instance('CWSource')
    events = ('before_delete_entity',)
    def __call__(self):
        if self.entity.name == 'system':
            msg = _("You cannot remove the system source")
            raise validation_error(self.entity, {None: msg})
        SourceRemovedOp(self._cw, uri=self.entity.name)


class SourceConfigUpdatedOp(hook.DataOperationMixIn, hook.Operation):

    def precommit_event(self):
        self.__processed = []
        for source in self.get_data():
            if not self.cnx.deleted_in_transaction(source.eid):
                conf = source.repo_source.check_config(source)
                self.__processed.append( (source, conf) )

    def postcommit_event(self):
        for source, conf in self.__processed:
            source.repo_source.update_config(source, conf)


class SourceRenamedOp(hook.LateOperation):
    oldname = newname = None # make pylint happy

    def postcommit_event(self):
        repo = self.cnx.repo
        # XXX race condition
        source = repo.sources_by_uri.pop(self.oldname)
        source.uri = self.newname
        source.public_config['uri'] = self.newname
        repo.sources_by_uri[self.newname] = source
        clear_cache(repo, 'source_defs')


class SourceUpdatedHook(SourceHook):
    __regid__ = 'cw.sources.configupdate'
    __select__ = SourceHook.__select__ & is_instance('CWSource')
    events = ('before_update_entity',)
    def __call__(self):
        if 'name' in self.entity.cw_edited:
            oldname, newname = self.entity.cw_edited.oldnewvalue('name')
            if oldname == 'system':
                msg = _("You cannot rename the system source")
                raise validation_error(self.entity, {('name', 'subject'): msg})
            SourceRenamedOp(self._cw, oldname=oldname, newname=newname)
        if 'config' in self.entity.cw_edited or 'url' in self.entity.cw_edited:
            if self.entity.name == 'system' and self.entity.config:
                msg = _("Configuration of the system source goes to "
                        "the 'sources' file, not in the database")
                raise validation_error(self.entity, {('config', 'subject'): msg})
            SourceConfigUpdatedOp.get_instance(self._cw).add_data(self.entity)


class SourceHostConfigUpdatedHook(SourceHook):
    __regid__ = 'cw.sources.hostconfigupdate'
    __select__ = SourceHook.__select__ & is_instance('CWSourceHostConfig')
    events = ('after_add_entity', 'after_update_entity', 'before_delete_entity',)
    def __call__(self):
        if self.entity.match(gethostname()):
            if self.event == 'after_update_entity' and \
                   not 'config' in self.entity.cw_edited:
                return
            try:
                SourceConfigUpdatedOp.get_instance(self._cw).add_data(self.entity.cwsource)
            except IndexError:
                # XXX no source linked to the host config yet
                pass



# coding: utf-8
# copyright 2003-2016 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of CubicWeb.
#
# CubicWeb is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# CubicWeb is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with CubicWeb.  If not, see <http://www.gnu.org/licenses/>.
"""unit tests for module cubicweb.utils"""

from six import string_types
from six.moves import cPickle as pickle
from six.moves.urllib.parse import urlsplit

from rql import parse

from logilab.common.testlib import TestCase, unittest_main, mock_object

from cubicweb.devtools.testlib import CubicWebTC
from cubicweb.rset import NotAnEntity, ResultSet, attr_desc_iterator
from cubicweb import NoResultError, MultipleResultsError


def pprelcachedict(d):
    res = {}
    for k, (rset, related) in d.items():
        res[k] = sorted(v.eid for v in related)
    return sorted(res.items())


class AttrDescIteratorTC(TestCase):
    """TestCase for cubicweb.rset.attr_desc_iterator"""

    def test_relations_description(self):
        """tests relations_description() function"""
        queries = {
            'Any U,L,M where U is CWUser, U login L, U mail M': [(1, 'login', 'subject'), (2, 'mail', 'subject')],
            'Any U,L,M where U is CWUser, L is Foo, U mail M': [(2, 'mail', 'subject')],
            'Any C,P where C is Company, C employs P': [(1, 'employs', 'subject')],
            'Any C,P where C is Company, P employed_by P': [],
            'Any C where C is Company, C employs P': [],
        }
        for rql, relations in queries.items():
            result = list(attr_desc_iterator(parse(rql).children[0], 0, 0))
            self.assertEqual((rql, result), (rql, relations))

    def test_relations_description_indexed(self):
        """tests relations_description() function"""
        queries = {
            'Any C,U,P,L,M where C is Company, C employs P, U is CWUser, U login L, U mail M':
            {0: [(2, 'employs', 'subject')],
             1: [(3, 'login', 'subject'), (4, 'mail', 'subject')]},
        }
        for rql, results in queries.items():
            for idx, relations in results.items():
                result = list(attr_desc_iterator(parse(rql).children[0], idx, idx))
                self.assertEqual(result, relations)

    def test_subquery_callfunc(self):
        rql = ('Any A,B,C,COUNT(D) GROUPBY A,B,C WITH A,B,C,D BEING '
               '(Any YEAR(CD), MONTH(CD), S, X WHERE X is CWUser, X creation_date CD, X in_state S)')
        rqlst = parse(rql)
        select, col = rqlst.locate_subquery(2, 'CWUser', None)
        result = list(attr_desc_iterator(select, col, 2))
        self.assertEqual(result, [])

    def test_subquery_callfunc_2(self):
        rql = ('Any X,S,L WHERE X in_state S WITH X, L BEING (Any X,MAX(L) GROUPBY X WHERE X is CWUser, T wf_info_for X, T creation_date L)')
        rqlst = parse(rql)
        select, col = rqlst.locate_subquery(0, 'CWUser', None)
        result = list(attr_desc_iterator(select, col, 0))
        self.assertEqual(result, [(1, 'in_state', 'subject')])


class ResultSetTC(CubicWebTC):

    def setUp(self):
        super(ResultSetTC, self).setUp()
        self.rset = ResultSet([[12, 'adim'], [13, 'syt']],
                              'Any U,L where U is CWUser, U login L',
                              description=[['CWUser', 'String'], ['Bar', 'String']])
        self.rset.req = mock_object(vreg=self.vreg, repo=self.repo)

    def compare_urls(self, url1, url2):
        info1 = urlsplit(url1)
        info2 = urlsplit(url2)
        self.assertEqual(info1[:3], info2[:3])
        if info1[3] != info2[3]:
            params1 = dict(pair.split('=') for pair in info1[3].split('&'))
            params2 = dict(pair.split('=') for pair in info1[3].split('&'))
            self.assertDictEqual(params1, params2)

    def test_pickle(self):
        del self.rset.req
        rs2 = pickle.loads(pickle.dumps(self.rset))
        self.assertEqual(self.rset.rows, rs2.rows)
        self.assertEqual(self.rset.rowcount, rs2.rowcount)
        self.assertEqual(self.rset.rql, rs2.rql)
        self.assertEqual(self.rset.description, rs2.description)

    def test_build_url(self):
        with self.admin_access.web_request() as req:
            baseurl = req.base_url()
            self.compare_urls(req.build_url('view', vid='foo', rql='yo'),
                              '%sview?vid=foo&rql=yo' % baseurl)
            self.compare_urls(req.build_url('view', _restpath='task/title/go'),
                              '%stask/title/go' % baseurl)
            #self.compare_urls(req.build_url('view', _restpath='/task/title/go'),
            #                  '%stask/title/go' % baseurl)
            # empty _restpath should not crash
            self.compare_urls(req.build_url('view', _restpath=''), baseurl)


    def test_build(self):
        """test basic build of a ResultSet"""
        rs = ResultSet([1,2,3], 'CWGroup X', description=['CWGroup', 'CWGroup', 'CWGroup'])
        self.assertEqual(rs.rowcount, 3)
        self.assertEqual(rs.rows, [1,2,3])
        self.assertEqual(rs.description, ['CWGroup', 'CWGroup', 'CWGroup'])


    def test_limit(self):
        rs = ResultSet([[12000, 'adim'], [13000, 'syt'], [14000, 'nico']],
                       'Any U,L where U is CWUser, U login L',
                       description=[['CWUser', 'String']] * 3)
        with self.admin_access.web_request() as req:
            rs.req = req
            rs.vreg = self.vreg
            self.assertEqual(rs.limit(2).rows, [[12000, 'adim'], [13000, 'syt']])
            rs2 = rs.limit(2, offset=1)
            self.assertEqual(rs2.rows, [[13000, 'syt'], [14000, 'nico']])
            self.assertEqual(rs2.get_entity(0, 0).cw_row, 0)
            self.assertEqual(rs.limit(2, offset=2).rows, [[14000, 'nico']])
            self.assertEqual(rs.limit(2, offset=3).rows, [])

    def test_limit_2(self):
        with self.admin_access.web_request() as req:
            rs = req.execute('Any E,U WHERE E is CWEType, E created_by U')
            # get entity on row 9. This will fill its created_by relation cache,
            # with cwuser on row 9 as well
            e1 = rs.get_entity(9, 0)
            # get entity on row 10. This will fill its created_by relation cache,
            # with cwuser built on row 9
            e2 = rs.get_entity(10, 0)
            # limit result set from row 10
            rs.limit(1, 10, inplace=True)
            # get back eid
            e = rs.get_entity(0, 0)
            self.assertTrue(e2 is e)
            # rs.limit has properly removed cwuser for request cache, but it's
            # still referenced by e/e2 relation cache
            u = e.created_by[0]
            # now ensure this doesn't trigger IndexError because cwuser.cw_row is 9
            # while now rset has only one row
            u.cw_rset[u.cw_row]

    def test_filter(self):
        rs = ResultSet([[12000, 'adim'], [13000, 'syt'], [14000, 'nico']],
                       'Any U,L where U is CWUser, U login L',
                       description=[['CWUser', 'String']] * 3)
        with self.admin_access.web_request() as req:
            rs.req = req
            rs.vreg = self.vreg
            def test_filter(entity):
                return entity.login != 'nico'

            rs2 = rs.filtered_rset(test_filter)
            self.assertEqual(len(rs2), 2)
            self.assertEqual([login for _, login in rs2], ['adim', 'syt'])
            self.assertEqual(rs2.description, rs.description[1:])

    def test_transform(self):
        rs = ResultSet([[12, 'adim'], [13, 'syt'], [14, 'nico']],
                       'Any U,L where U is CWUser, U login L',
                       description=[['CWUser', 'String']] * 3)
        with self.admin_access.web_request() as req:
            rs.req = req
            def test_transform(row, desc):
                return row[1:], desc[1:]
            rs2 = rs.transformed_rset(test_transform)

            self.assertEqual(len(rs2), 3)
            self.assertEqual(list(rs2), [['adim'],['syt'],['nico']])

    def test_sort(self):
        rs = ResultSet([[12000, 'adim'], [13000, 'syt'], [14000, 'nico']],
                       'Any U,L where U is CWUser, U login L',
                       description=[['CWUser', 'String']] * 3)
        with self.admin_access.web_request() as req:
            rs.req = req
            rs.vreg = self.vreg

            rs2 = rs.sorted_rset(lambda e:e.cw_attr_cache['login'])
            self.assertEqual(len(rs2), 3)
            self.assertEqual([login for _, login in rs2], ['adim', 'nico', 'syt'])
            # make sure rs is unchanged
            self.assertEqual([login for _, login in rs], ['adim', 'syt', 'nico'])

            rs2 = rs.sorted_rset(lambda e:e.cw_attr_cache['login'], reverse=True)
            self.assertEqual(len(rs2), 3)
            self.assertEqual([login for _, login in rs2], ['syt', 'nico', 'adim'])
            # make sure rs is unchanged
            self.assertEqual([login for _, login in rs], ['adim', 'syt', 'nico'])

            rs3 = rs.sorted_rset(lambda row: row[1], col=-1)
            self.assertEqual(len(rs3), 3)
            self.assertEqual([login for _, login in rs3], ['adim', 'nico', 'syt'])
            # make sure rs is unchanged
            self.assertEqual([login for _, login in rs], ['adim', 'syt', 'nico'])

    def test_split(self):
        rs = ResultSet([[12000, 'adim', u'Adim chez les pinguins'],
                        [12000, 'adim', u'Jardiner facile'],
                        [13000, 'syt',  u'Le carrelage en 42 leçons'],
                        [14000, 'nico', u'La tarte tatin en 15 minutes'],
                        [14000, 'nico', u"L'épluchage du castor commun"]],
                       'Any U, L, T WHERE U is CWUser, U login L,'\
                       'D created_by U, D title T',
                       description=[['CWUser', 'String', 'String']] * 5)
        with self.admin_access.web_request() as req:
            rs.req = req
            rs.vreg = self.vreg
            rsets = rs.split_rset(lambda e:e.cw_attr_cache['login'])
            self.assertEqual(len(rsets), 3)
            self.assertEqual([login for _, login,_ in rsets[0]], ['adim', 'adim'])
            self.assertEqual([login for _, login,_ in rsets[1]], ['syt'])
            self.assertEqual([login for _, login,_ in rsets[2]], ['nico', 'nico'])
            # make sure rs is unchanged
            self.assertEqual([login for _, login,_ in rs], ['adim', 'adim', 'syt', 'nico', 'nico'])

            rsets = rs.split_rset(lambda e:e.cw_attr_cache['login'], return_dict=True)
            self.assertEqual(len(rsets), 3)
            self.assertEqual([login for _, login,_ in rsets['nico']], ['nico', 'nico'])
            self.assertEqual([login for _, login,_ in rsets['adim']], ['adim', 'adim'])
            self.assertEqual([login for _, login,_ in rsets['syt']], ['syt'])
            # make sure rs is unchanged
            self.assertEqual([login for _, login,_ in rs], ['adim', 'adim', 'syt', 'nico', 'nico'])

            rsets = rs.split_rset(lambda s: s.count('d'), col=2)
            self.assertEqual(len(rsets), 2)
            self.assertEqual([title for _, _, title in rsets[0]],
                              [u"Adim chez les pinguins",
                               u"Jardiner facile",
                               u"L'épluchage du castor commun",])
            self.assertEqual([title for _, _, title in rsets[1]],
                              [u"Le carrelage en 42 leçons",
                               u"La tarte tatin en 15 minutes",])
            # make sure rs is unchanged
            self.assertEqual([title for _, _, title in rs],
                              [u'Adim chez les pinguins',
                               u'Jardiner facile',
                               u'Le carrelage en 42 leçons',
                               u'La tarte tatin en 15 minutes',
                               u"L'épluchage du castor commun"])

    def test_cached_syntax_tree(self):
        """make sure syntax tree is cached"""
        rqlst1 = self.rset.syntax_tree()
        rqlst2 = self.rset.syntax_tree()
        self.assertIs(rqlst1, rqlst2)

    def test_get_entity_simple(self):
        with self.admin_access.web_request() as req:
            req.create_entity('CWUser', login=u'adim', upassword='adim',
                              surname=u'di mascio', firstname=u'adrien')
            req.drop_entity_cache()
            e = req.execute('Any X,T WHERE X login "adim", X surname T').get_entity(0, 0)
            self.assertEqual(e.cw_attr_cache['surname'], 'di mascio')
            self.assertRaises(KeyError, e.cw_attr_cache.__getitem__, 'firstname')
            self.assertRaises(KeyError, e.cw_attr_cache.__getitem__, 'creation_date')
            self.assertEqual(pprelcachedict(e._cw_related_cache), [])
            e.complete()
            self.assertEqual(e.cw_attr_cache['firstname'], 'adrien')
            self.assertEqual(pprelcachedict(e._cw_related_cache), [])

    def test_get_entity_advanced(self):
        with self.admin_access.web_request() as req:
            req.create_entity('Bookmark', title=u'zou', path=u'/view')
            req.drop_entity_cache()
            req.execute('SET X bookmarked_by Y WHERE X is Bookmark, Y login "anon"')
            rset = req.execute('Any X,Y,XT,YN WHERE X bookmarked_by Y, X title XT, Y login YN')

            e = rset.get_entity(0, 0)
            self.assertEqual(e.cw_row, 0)
            self.assertEqual(e.cw_col, 0)
            self.assertEqual(e.cw_attr_cache['title'], 'zou')
            self.assertRaises(KeyError, e.cw_attr_cache.__getitem__, 'path')
            other_rset = req.execute('Any X, P WHERE X is Bookmark, X path P')
            # check that get_entity fetches e from the request's cache, and
            # updates it with attributes from the new rset
            self.assertIs(other_rset.get_entity(0, 0), e)
            self.assertIn('path', e.cw_attr_cache)
            self.assertEqual(e.view('text'), 'zou')
            self.assertEqual(pprelcachedict(e._cw_related_cache), [])

            e = rset.get_entity(0, 1)
            self.assertEqual(e.cw_row, 0)
            self.assertEqual(e.cw_col, 1)
            self.assertEqual(e.cw_attr_cache['login'], 'anon')
            self.assertRaises(KeyError, e.cw_attr_cache.__getitem__, 'firstname')
            self.assertEqual(pprelcachedict(e._cw_related_cache),
                              [])
            e.complete()
            self.assertEqual(e.cw_attr_cache['firstname'], None)
            self.assertEqual(e.view('text'), 'anon')
            self.assertEqual(pprelcachedict(e._cw_related_cache),
                              [])

            self.assertRaises(NotAnEntity, rset.get_entity, 0, 2)
            self.assertRaises(NotAnEntity, rset.get_entity, 0, 3)

    def test_get_entity_relation_cache_compt(self):
        with self.admin_access.web_request() as req:
            rset = req.execute('Any X,S WHERE X in_state S, X login "anon"')
            e = rset.get_entity(0, 0)
            seid = req.execute('State X WHERE X name "activated"')[0][0]
            # for_user / in_group are prefetched in CWUser __init__, in_state should
            # be filed from our query rset
            self.assertEqual(pprelcachedict(e._cw_related_cache),
                              [('in_state_subject', [seid])])

    def test_get_entity_advanced_prefilled_cache(self):
        with self.admin_access.web_request() as req:
            e = req.create_entity('Bookmark', title=u'zou', path=u'path')
            req.cnx.commit()
            rset = req.execute('Any X,U,S,XT,UL,SN WHERE X created_by U, U in_state S, '
                                'X title XT, S name SN, U login UL, X eid %s' % e.eid)
            e = rset.get_entity(0, 0)
            self.assertEqual(e.cw_attr_cache['title'], 'zou')
            self.assertEqual(pprelcachedict(e._cw_related_cache),
                             [('created_by_subject', [req.user.eid])])
            # first level of recursion
            u = e.created_by[0]
            self.assertEqual(u.cw_attr_cache['login'], 'admin')
            self.assertRaises(KeyError, u.cw_attr_cache.__getitem__, 'firstname')
            # second level of recursion
            s = u.in_state[0]
            self.assertEqual(s.cw_attr_cache['name'], 'activated')
            self.assertRaises(KeyError, s.cw_attr_cache.__getitem__, 'description')

    def test_get_entity_recursion(self):
        with self.admin_access.repo_cnx() as cnx:
            cnx.create_entity('EmailAddress', address=u'toto',
                              reverse_primary_email=cnx.user.eid)
            cnx.commit()

        # get_entity should fill the caches for user and email, even if both
        # entities are already in the connection's entity cache
        with self.admin_access.repo_cnx() as cnx:
            mail = cnx.find('EmailAddress').one()
            rset = cnx.execute('Any X, E WHERE X primary_email E')
            u = rset.get_entity(0, 0)
            self.assertTrue(u.cw_relation_cached('primary_email', 'subject'))
            self.assertTrue(mail.cw_relation_cached('primary_email', 'object'))

        with self.admin_access.repo_cnx() as cnx:
            mail = cnx.find('EmailAddress').one()
            rset = cnx.execute('Any X, E WHERE X primary_email E')
            rset.get_entity(0, 1)
            self.assertTrue(mail.cw_relation_cached('primary_email', 'object'))
            u = cnx.user
            self.assertTrue(u.cw_relation_cached('primary_email', 'subject'))


    def test_get_entity_cache_with_left_outer_join(self):
        with self.admin_access.web_request() as req:
            eid = req.execute('INSERT CWUser E: E login "joe", E upassword "joe", E in_group G '
                               'WHERE G name "users"')[0][0]
            rset = req.execute('Any X,E WHERE X eid %(x)s, X primary_email E?', {'x': eid})
            e = rset.get_entity(0, 0)
            # if any of the assertion below fails with a KeyError, the relation is not cached
            # related entities should be an empty list
            self.assertEqual(e._cw_related_cache['primary_email_subject'][True], ())
            # related rset should be an empty rset
            cached = e._cw_related_cache['primary_email_subject'][False]
            self.assertIsInstance(cached, ResultSet)
            self.assertEqual(cached.rowcount, 0)


    def test_get_entity_union(self):
        with self.admin_access.web_request() as req:
            e = req.create_entity('Bookmark', title=u'manger', path=u'path')
            req.drop_entity_cache()
            rset = req.execute('Any X,N ORDERBY N WITH X,N BEING '
                                '((Any X,N WHERE X is Bookmark, X title N)'
                                ' UNION '
                                ' (Any X,N WHERE X is CWGroup, X name N))')
            expected = (('CWGroup', 'guests'), ('CWGroup', 'managers'),
                        ('Bookmark', 'manger'), ('CWGroup', 'owners'),
                        ('CWGroup', 'users'))
            for entity in rset.entities(): # test get_entity for each row actually
                etype, n = expected[entity.cw_row]
                self.assertEqual(entity.cw_etype, etype)
                attr = etype == 'Bookmark' and 'title' or 'name'
                self.assertEqual(entity.cw_attr_cache[attr], n)

    def test_one(self):
        with self.admin_access.web_request() as req:
            req.create_entity('CWUser', login=u'cdevienne',
                                         upassword=u'cdevienne',
                                         surname=u'de Vienne',
                                         firstname=u'Christophe')
            e = req.execute('Any X WHERE X login "cdevienne"').one()

            self.assertEqual(e.surname, u'de Vienne')

            e = req.execute(
                'Any X, N WHERE X login "cdevienne", X surname N').one()
            self.assertEqual(e.surname, u'de Vienne')

            e = req.execute(
                'Any N, X WHERE X login "cdevienne", X surname N').one(col=1)
            self.assertEqual(e.surname, u'de Vienne')

    def test_one_no_rows(self):
        with self.admin_access.web_request() as req:
            with self.assertRaises(NoResultError):
                req.execute('Any X WHERE X login "patanok"').one()

    def test_one_multiple_rows(self):
        with self.admin_access.web_request() as req:
            req.create_entity(
                'CWUser', login=u'cdevienne', upassword=u'cdevienne',
                surname=u'de Vienne', firstname=u'Christophe')

            req.create_entity(
                'CWUser', login=u'adim', upassword='adim', surname=u'di mascio',
                firstname=u'adrien')

            with self.assertRaises(MultipleResultsError):
                req.execute('Any X WHERE X is CWUser').one()

    def test_related_entity_optional(self):
        with self.admin_access.web_request() as req:
            e = req.create_entity('Bookmark', title=u'aaaa', path=u'path')
            rset = req.execute('Any B,U,L WHERE B bookmarked_by U?, U login L')
            entity, rtype = rset.related_entity(0, 2)
            self.assertEqual(entity, None)
            self.assertEqual(rtype, None)

    def test_related_entity_union_subquery_1(self):
        with self.admin_access.web_request() as req:
            e = req.create_entity('Bookmark', title=u'aaaa', path=u'path')
            rset = req.execute('Any X,N ORDERBY N WITH X,N BEING '
                                '((Any X,N WHERE X is CWGroup, X name N)'
                                ' UNION '
                                ' (Any X,N WHERE X is Bookmark, X title N))')
            entity, rtype = rset.related_entity(0, 1)
            self.assertEqual(entity.eid, e.eid)
            self.assertEqual(rtype, 'title')
            self.assertEqual(entity.title, 'aaaa')
            entity, rtype = rset.related_entity(1, 1)
            self.assertEqual(entity.cw_etype, 'CWGroup')
            self.assertEqual(rtype, 'name')
            self.assertEqual(entity.name, 'guests')

    def test_related_entity_union_subquery_2(self):
        with self.admin_access.web_request() as req:
            e = req.create_entity('Bookmark', title=u'aaaa', path=u'path')
            rset = req.execute('Any X,N ORDERBY N WHERE X is Bookmark WITH X,N BEING '
                                '((Any X,N WHERE X is CWGroup, X name N)'
                                ' UNION '
                                ' (Any X,N WHERE X is Bookmark, X title N))')
            entity, rtype = rset.related_entity(0, 1)
            self.assertEqual(entity.eid, e.eid)
            self.assertEqual(rtype, 'title')
            self.assertEqual(entity.title, 'aaaa')

    def test_related_entity_union_subquery_3(self):
        with self.admin_access.web_request() as req:
            e = req.create_entity('Bookmark', title=u'aaaa', path=u'path')
            rset = req.execute('Any X,N ORDERBY N WITH N,X BEING '
                                '((Any N,X WHERE X is CWGroup, X name N)'
                                ' UNION '
                                ' (Any N,X WHERE X is Bookmark, X title N))')
            entity, rtype = rset.related_entity(0, 1)
            self.assertEqual(entity.eid, e.eid)
            self.assertEqual(rtype, 'title')
            self.assertEqual(entity.title, 'aaaa')

    def test_related_entity_union_subquery_4(self):
        with self.admin_access.web_request() as req:
            e = req.create_entity('Bookmark', title=u'aaaa', path=u'path')
            rset = req.execute('Any X,X, N ORDERBY N WITH X,N BEING '
                                '((Any X,N WHERE X is CWGroup, X name N)'
                                ' UNION '
                                ' (Any X,N WHERE X is Bookmark, X title N))')
            entity, rtype = rset.related_entity(0, 2)
            self.assertEqual(entity.eid, e.eid)
            self.assertEqual(rtype, 'title')
            self.assertEqual(entity.title, 'aaaa')

    def test_related_entity_trap_subquery(self):
        with self.admin_access.web_request() as req:
            req.create_entity('Bookmark', title=u'test bookmark', path=u'')
            req.execute('SET B bookmarked_by U WHERE U login "admin"')
            rset = req.execute('Any B,T,L WHERE B bookmarked_by U, U login L '
                                'WITH B,T BEING (Any B,T WHERE B is Bookmark, B title T)')
            rset.related_entity(0, 2)

    def test_related_entity_subquery_outerjoin(self):
        with self.admin_access.web_request() as req:
            rset = req.execute('Any X,S,L WHERE X in_state S '
                                'WITH X, L BEING (Any X,MAX(L) GROUPBY X '
                                'WHERE X is CWUser, T? wf_info_for X, T creation_date L)')
            self.assertEqual(len(rset), 2)
            rset.related_entity(0, 1)
            rset.related_entity(0, 2)

    def test_entities(self):
        with self.admin_access.web_request() as req:
            rset = req.execute('Any U,G WHERE U in_group G')
            # make sure we have at least one element
            self.assertTrue(rset)
            self.assertEqual(set(e.e_schema.type for e in rset.entities(0)),
                              set(['CWUser',]))
            self.assertEqual(set(e.e_schema.type for e in rset.entities(1)),
                              set(['CWGroup',]))

    def test_iter_rows_with_entities(self):
        with self.admin_access.web_request() as req:
            rset = req.execute('Any U,UN,G,GN WHERE U in_group G, U login UN, G name GN')
            # make sure we have at least one element
            self.assertTrue(rset)
            out = list(rset.iter_rows_with_entities())[0]
            self.assertEqual( out[0].login, out[1] )
            self.assertEqual( out[2].name, out[3] )

    def test_printable_rql(self):
        with self.admin_access.web_request() as req:
            rset = req.execute(u'CWEType X WHERE X final FALSE')
            self.assertEqual(rset.printable_rql(),
                              'Any X WHERE X final FALSE, X is CWEType')

    def test_searched_text(self):
        with self.admin_access.web_request() as req:
            rset = req.execute(u'Any X WHERE X has_text "foobar"')
            self.assertEqual(rset.searched_text(), 'foobar')
            rset = req.execute(u'Any X WHERE X has_text %(text)s', {'text' : 'foo'})
            self.assertEqual(rset.searched_text(), 'foo')

    def test_union_limited_rql(self):
        with self.admin_access.web_request() as req:
            rset = req.execute('(Any X,N WHERE X is Bookmark, X title N)'
                                ' UNION '
                                '(Any X,N WHERE X is CWGroup, X name N)')
            rset.limit(2, 10, inplace=True)
            self.assertEqual(rset.limited_rql(),
                              'Any A,B LIMIT 2 OFFSET 10 '
                              'WITH A,B BEING ('
                              '(Any X,N WHERE X is Bookmark, X title N) '
                              'UNION '
                              '(Any X,N WHERE X is CWGroup, X name N)'
                              ')')

    def test_possible_actions_cache(self):
        with self.admin_access.web_request() as req:
            rset = req.execute('Any D, COUNT(U) GROUPBY D WHERE U is CWUser, U creation_date D')
            rset.possible_actions(argument='Value')
            self.assertRaises(AssertionError, rset.possible_actions, argument='OtherValue')
            self.assertRaises(AssertionError, rset.possible_actions, other_argument='Value')

    def test_count_users_by_date(self):
        with self.admin_access.web_request() as req:
            rset = req.execute('Any D, COUNT(U) GROUPBY D WHERE U is CWUser, U creation_date D')
            self.assertEqual(rset.related_entity(0,0), (None, None))

    def test_str(self):
        with self.admin_access.web_request() as req:
            rset = req.execute('(Any X,N WHERE X is CWGroup, X name N)')
            self.assertIsInstance(str(rset), string_types)
            self.assertEqual(len(str(rset).splitlines()), 1)

    def test_repr(self):
        with self.admin_access.web_request() as req:
            rset = req.execute('(Any X,N WHERE X is CWGroup, X name N)')
            self.assertIsInstance(repr(rset), string_types)
            self.assertTrue(len(repr(rset).splitlines()) > 1)

            rset = req.execute('(Any X WHERE X is CWGroup, X name "managers")')
            self.assertIsInstance(str(rset), string_types)
            self.assertEqual(len(str(rset).splitlines()), 1)

    def test_slice(self):
        rs = ResultSet([[12000, 'adim', u'Adim chez les pinguins'],
                        [12000, 'adim', u'Jardiner facile'],
                        [13000, 'syt',  u'Le carrelage en 42 leçons'],
                        [14000, 'nico', u'La tarte tatin en 15 minutes'],
                        [14000, 'nico', u"L'épluchage du castor commun"]],
                       'Any U, L, T WHERE U is CWUser, U login L,'\
                       'D created_by U, D title T',
                       description=[['CWUser', 'String', 'String']] * 5)
        self.assertEqual(rs[1::2],
            [[12000, 'adim', u'Jardiner facile'],
             [14000, 'nico', u'La tarte tatin en 15 minutes']])

    def test_nonregr_symmetric_relation(self):
        # see https://www.cubicweb.org/ticket/4739253
        with self.admin_access.client_cnx() as cnx:
            p1 = cnx.create_entity('Personne', nom=u'sylvain')
            cnx.create_entity('Personne', nom=u'denis', connait=p1)
            cnx.commit()
            rset = cnx.execute('Any X,Y WHERE X connait Y')
            rset.get_entity(0, 1) # used to raise KeyError

if __name__ == '__main__':
    unittest_main()



# -*- coding: utf-8 -*-
# copyright 2003-2016 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of CubicWeb.
#
# CubicWeb is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# CubicWeb is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with CubicWeb.  If not, see <http://www.gnu.org/licenses/>.
"""functional tests for core hooks

Note:
  syncschema.py hooks are mostly tested in server/test/unittest_migrations.py
"""

from six import text_type

from cubicweb import ValidationError
from cubicweb.devtools.testlib import CubicWebTC


class CWPropertyHooksTC(CubicWebTC):

    def test_unexistant_cwproperty(self):
        with self.admin_access.web_request() as req:
            with self.assertRaises(ValidationError) as cm:
                req.execute('INSERT CWProperty X: X pkey "bla.bla", '
                            'X value "hop", X for_user U')
            cm.exception.translate(text_type)
            self.assertEqual(cm.exception.errors,
                             {'pkey-subject': 'unknown property key bla.bla'})

            with self.assertRaises(ValidationError) as cm:
                req.execute('INSERT CWProperty X: X pkey "bla.bla", X value "hop"')
            cm.exception.translate(text_type)
            self.assertEqual(cm.exception.errors,
                             {'pkey-subject': 'unknown property key bla.bla'})

    def test_site_wide_cwproperty(self):
        with self.admin_access.web_request() as req:
            with self.assertRaises(ValidationError) as cm:
                req.execute('INSERT CWProperty X: X pkey "ui.site-title", '
                            'X value "hop", X for_user U')
            self.assertEqual(cm.exception.errors,
                             {'for_user-subject': "site-wide property can't be set for user"})

    def test_system_cwproperty(self):
        with self.admin_access.web_request() as req:
            with self.assertRaises(ValidationError) as cm:
                req.execute('INSERT CWProperty X: X pkey "system.version.cubicweb", '
                            'X value "hop", X for_user U')
            self.assertEqual(cm.exception.errors,
                             {'for_user-subject': "site-wide property can't be set for user"})

    def test_bad_type_cwproperty(self):
        with self.admin_access.web_request() as req:
            with self.assertRaises(ValidationError) as cm:
                req.execute('INSERT CWProperty X: X pkey "ui.language", '
                            'X value "hop", X for_user U')
            self.assertEqual(cm.exception.errors,
                             {'value-subject': u'unauthorized value'})
            with self.assertRaises(ValidationError) as cm:
                req.execute('INSERT CWProperty X: X pkey "ui.language", X value "hop"')
            self.assertEqual(cm.exception.errors, {'value-subject': u'unauthorized value'})

    def test_vreg_propertyvalues_update(self):
        self.vreg.register_property(
            'test.int', type='Int', help='', sitewide=True)
        with self.admin_access.repo_cnx() as cnx:
            cnx.execute('INSERT CWProperty X: X pkey "test.int", X value "42"')
            cnx.commit()
        self.assertEqual(self.vreg.property_value('test.int'), 42)

    def test_sync_user_props(self):
        with self.admin_access.client_cnx() as cnx:
            self.assertNotIn('ui.language', cnx.user.properties)
            cnx.user.set_property(u'ui.language', u'fr')
            self.assertNotIn('ui.language', cnx.user.properties)
            cnx.commit()
            self.assertEqual(cnx.user.properties['ui.language'], 'fr')
            cnx.user.set_property(u'ui.language', u'en')
            self.assertEqual(cnx.user.properties['ui.language'], 'fr')
            cnx.commit()
            self.assertEqual(cnx.user.properties['ui.language'], 'en')
            cnx.execute('DELETE CWProperty X WHERE X for_user U, U eid %(u)s',
                        {'u': cnx.user.eid})
            self.assertEqual(cnx.user.properties['ui.language'], 'en')
            cnx.commit()
            self.assertNotIn('ui.language', cnx.user.properties)

    def test_sync_sitewide_props(self):
        with self.admin_access.client_cnx() as cnx:
            self.assertNotIn('ui.language', cnx.vreg['propertyvalues'])
            cwprop = cnx.create_entity('CWProperty', pkey=u'ui.language', value=u'fr')
            self.assertNotIn('ui.language', cnx.vreg['propertyvalues'])
            cnx.commit()
            self.assertEqual(cnx.vreg['propertyvalues']['ui.language'], 'fr')
            cwprop.cw_set(value=u'en')
            self.assertEqual(cnx.vreg['propertyvalues']['ui.language'], 'fr')
            cnx.commit()
            self.assertEqual(cnx.vreg['propertyvalues']['ui.language'], 'en')
            cwprop.cw_delete()
            self.assertEqual(cnx.vreg['propertyvalues']['ui.language'], 'en')
            cnx.commit()
            self.assertNotIn('ui.language', cnx.vreg['propertyvalues'])


class UserGroupsSyncTC(CubicWebTC):

    def test_sync_groups(self):
        with self.admin_access.client_cnx() as cnx:
            cnx.execute('SET U in_group G WHERE G name "users", U eid %(u)s',
                        {'u': cnx.user.eid})
            self.assertEqual(cnx.user.groups, set(['managers']))
            cnx.commit()
            self.assertEqual(cnx.user.groups, set(['managers', 'users']))
            cnx.execute('DELETE U in_group G WHERE G name "users", U eid %(u)s',
                        {'u': cnx.user.eid})
            self.assertEqual(cnx.user.groups, set(['managers', 'users']))
            cnx.commit()
            self.assertEqual(cnx.user.groups, set(['managers']))


if __name__ == '__main__':
    import unittest
    unittest.main()
